""" BlueSky traffic implementation."""
from collections.abc import Collection
from math import *
from random import randint
import numpy as np

import bluesky as bs
from bluesky.core import Entity, Timer
from bluesky.stack.recorder import savecmd
from bluesky.tools import geo
from bluesky.tools.misc import latlon2txt
from bluesky.tools.aero import casormach2tas, fpm, kts, ft, g0, Rearth, nm, tas2cas,\
                         vatmos,  vtas2cas, vtas2mach, vcasormach


from bluesky.traffic.asas import ConflictDetection, ConflictResolution
from .windsim import WindSim
from .conditional import Condition
from .trails import Trails
from .adsbmodel import ADSB
from .aporasas import APorASAS
from .autopilot import Autopilot
from .activewpdata import ActiveWaypoint
from .turbulence import Turbulence
from .trafficgroups import TrafficGroups
from .performance.perfbase import PerfBase

# Register settings defaults
bs.settings.set_variable_defaults(performance_model='openap', asas_dt=1.0)

# if bs.settings.performance_model == 'bada':
#     try:
#         print('Using BADA Performance model')
#         from .performance.bada.perfbada import PerfBADA as Perf
#     except Exception as err:# ImportError as err:
#         print(err)
#         print('Falling back to Open Aircraft Performance (OpenAP) model')
#         bs.settings.performance_model = "openap"
#         from .performance.openap import OpenAP as Perf
# elif bs.settings.performance_model == 'openap':
#     print('Using Open Aircraft Performance (OpenAP) model')
#     from .performance.openap import OpenAP as Perf
# else:
#     print('Using BlueSky legacy performance model')
#     from .performance.legacy.perfbs import PerfBS as Perf


class Traffic(Entity):
    """
    Traffic class definition    : Traffic data
    Methods:
        Traffic()            :  constructor
        reset()              :  Reset traffic database w.r.t a/c data
        create(acid,actype,aclat,aclon,achdg,acalt,acspd) : create aircraft
        delete(acid)         : delete an aircraft from traffic data
        deletall()           : delete all traffic
        update(sim)          : do a numerical integration step
        id2idx(name)         : return index in traffic database of given call sign
        engchange(i,engtype) : change engine type of an aircraft
        setnoise(A)          : Add turbulence
    Members: see create
    Created by  : Jacco M. Hoekstra
    """

    def __init__(self):
        super().__init__()

        # Traffic is the toplevel trafficarrays object
        self.setroot(self)

        self.ntraf = 0

        self.cond = Condition()  # Conditional commands list
        self.wind = WindSim()
        self.turbulence = Turbulence()
        self.translvl = 5000.*ft # [m] Default transition level

        # Default commands issued for an aircraft after creation
        self.crecmdlist = []

        # Manual timer for CD and CR
        self.asastimer = Timer(name='asas', dt=bs.settings.asas_dt)

        with self.settrafarrays():
            # Aircraft Info
            self.id      = []  # identifier (string)
            self.type    = []  # aircaft type (string)

            # Positions
            self.lat     = np.array([])  # latitude [deg]
            self.lon     = np.array([])  # longitude [deg]
            self.distflown = np.array([])  # distance travelled [m]
            self.alt     = np.array([])  # altitude [m]
            self.hdg     = np.array([])  # traffic heading [deg]
            self.trk     = np.array([])  # track angle [deg]

            # Velocities
            self.tas     = np.array([])  # true airspeed [m/s]
            self.gs      = np.array([])  # ground speed [m/s]
            self.gsnorth = np.array([])  # ground speed [m/s]
            self.gseast  = np.array([])  # ground speed [m/s]
            self.cas     = np.array([])  # calibrated airspeed [m/s]
            self.M       = np.array([])  # mach number
            self.vs      = np.array([])  # vertical speed [m/s]

            # Acceleration
            self.ax = np.array([])  # [m/s2] current longitudinal acceleration

            # Atmosphere
            self.p       = np.array([])  # air pressure [N/m2]
            self.rho     = np.array([])  # air density [kg/m3]
            self.Temp    = np.array([])  # air temperature [K]
            self.dtemp   = np.array([])  # delta t for non-ISA conditions

            # Wind speeds
            self.windnorth = np.array([])  # wind speed north component a/c pos [m/s]
            self.windeast  = np.array([])  # wind speed east component a/c pos [m/s]

            # Traffic autopilot settings
            self.selspd = np.array([])  # selected spd(CAS or Mach) [m/s or -]
            self.aptas  = np.array([])  # just for initializing
            self.selalt = np.array([])  # selected alt[m]
            self.selvs  = np.array([])  # selected vertical speed [m/s]

            # Whether to perform LNAV and VNAV
            self.swlnav    = np.array([], dtype=bool)
            self.swvnav    = np.array([], dtype=bool)
            self.swvnavspd = np.array([], dtype=bool)

            # Flight Models
            self.cd       = ConflictDetection()
            self.cr       = ConflictResolution()
            self.ap       = Autopilot()
            self.aporasas = APorASAS()
            self.adsb     = ADSB()
            self.trails   = Trails()
            self.actwp    = ActiveWaypoint()
            self.perf     = PerfBase()

            # Group Logic
            self.groups = TrafficGroups()

            # Traffic autopilot data
            self.swhdgsel = np.array([], dtype=bool)  # determines whether aircraft is turning

            # Traffic autothrottle settings
            self.swats    = np.array([], dtype=bool)  # Switch indicating whether autothrottle system is on/off
            self.thr      = np.array([])        # Thottle seeting (0.0-1.0), negative = non-valid/auto

            # Display information on label
            self.label       = []  # Text and bitmap of traffic label

            # Miscallaneous
            self.coslat = np.array([])  # Cosine of latitude for computations
            self.eps    = np.array([])  # Small nonzero numbers
            self.work   = np.array([])  # Work done throughout the flight

        # Default bank angles per flight phase
        self.bphase = np.deg2rad(np.array([15, 35, 35, 35, 15, 45]))

    def reset(self):
        ''' Clear all traffic data upon simulation reset. '''
        # Some child reset functions depend on a correct value of self.ntraf
        self.ntraf = 0
        # This ensures that the traffic arrays (which size is dynamic)
        # are all reset as well, so all lat,lon,sdp etc but also objects adsb
        super().reset()

        # reset performance model
        self.perf.reset()

        # Reset models
        self.wind.clear()

        # Build new modules for turbulence
        self.turbulence.reset()

        # Noise (turbulence, ADBS-transmission noise, ADSB-truncated effect)
        self.setnoise(False)

        # Reset transition level to default value
        self.translvl = 5000.*ft

    def mcre(self, n, actype="B744", acalt=None, acspd=None, dest=None):
        """ Create one or more random aircraft in a specified area """
        area = bs.ref.area.bbox

        # Generate random callsigns
        idtmp = chr(randint(65, 90)) + chr(randint(65, 90)) + '{:>05}'
        acid = [idtmp.format(i) for i in range(n)]

        # Generate random positions
        aclat = np.random.rand(n) * (area[2] - area[0]) + area[0]
        aclon = np.random.rand(n) * (area[3] - area[1]) + area[1]
        achdg = np.random.randint(1, 360, n)
        acalt = acalt or np.random.randint(2000, 39000, n) * ft
        acspd = acspd or np.random.randint(250, 450, n) * kts

        self.cre(acid, actype, aclat, aclon, achdg, acalt, acspd)


    def cre(self, acid, actype="B744", aclat=52., aclon=4., achdg=None, acalt=0, acspd=0):
        """ Create one or more aircraft. """
        # Determine number of aircraft to create from array length of acid
        n = 1 if isinstance(acid, str) else len(acid)

        if isinstance(acid, str):
            # Check if not already exist
            if self.id.count(acid.upper()) > 0:
                return False, acid + " already exists."  # already exists do nothing
            acid = n * [acid]

        # Adjust the size of all traffic arrays
        super().create(n)
        self.ntraf += n

        if isinstance(actype, str):
            actype = n * [actype]

        if isinstance(aclat, (float, int)):
            aclat = np.array(n * [aclat])

        if isinstance(aclon, (float, int)):
            aclon = np.array(n * [aclon])

        # Limit longitude to [-180.0, 180.0]
        aclon[aclon > 180.0] -= 360.0
        aclon[aclon < -180.0] += 360.0

        achdg = (bs.ref.hdg or 0.0) if achdg is None else achdg

        # Aircraft Info
        self.id[-n:]   = acid
        self.type[-n:] = actype

        # Positions
        self.lat[-n:]  = aclat
        self.lon[-n:]  = aclon
        self.alt[-n:]  = acalt

        self.hdg[-n:]  = achdg
        self.trk[-n:]  = achdg

        # Velocities
        self.tas[-n:], self.cas[-n:], self.M[-n:] = vcasormach(acspd, acalt)
        self.gs[-n:]      = self.tas[-n:]
        hdgrad = np.radians(achdg)
        self.gsnorth[-n:] = self.tas[-n:] * np.cos(hdgrad)
        self.gseast[-n:] = self.tas[-n:] * np.sin(hdgrad)

        # Atmosphere
        self.p[-n:], self.rho[-n:], self.Temp[-n:] = vatmos(acalt)

        # Wind
        if self.wind.winddim > 0:
            applywind         = self.alt[-n:]> 50.*ft
            self.windnorth[-n:], self.windeast[-n:]  = self.wind.getdata(self.lat[-n:], self.lon[-n:], self.alt[-n:])
            self.gsnorth[-n:] = self.gsnorth[-n:] + self.windnorth[-n:]*applywind
            self.gseast[-n:]  = self.gseast[-n:]  + self.windeast[-n:]*applywind
            self.trk[-n:]     = np.logical_not(applywind)*achdg + \
                                applywind*np.degrees(np.arctan2(self.gseast[-n:], self.gsnorth[-n:]))
            self.gs[-n:]      = np.sqrt(self.gsnorth[-n:]**2 + self.gseast[-n:]**2)
        else:
            self.windnorth[-n:] = 0.0
            self.windeast[-n:]  = 0.0

        # Traffic autopilot settings
        self.selspd[-n:] = self.cas[-n:]
        self.aptas[-n:]  = self.tas[-n:]
        self.selalt[-n:] = self.alt[-n:]

        # Display information on label
        self.label[-n:] = n*[['', '', '', 0]]

        # Miscallaneous
        self.coslat[-n:] = np.cos(np.radians(aclat))  # Cosine of latitude for flat-earth aproximations
        self.eps[-n:] = 0.01

        # Finally call create for child TrafficArrays. This only needs to be done
        # manually in Traffic.
        self.create_children(n)

        # Record as individual CRE commands for repeatability
        #print(self.ntraf-n,self.ntraf)
        for j in range(self.ntraf-n,self.ntraf):
            # Reconstruct CRE command
            line = "CRE "+",".join([self.id[j],self.type[j],
                                    str(self.lat[j]),str(self.lon[j]),
                                    str(round(self.trk[j])),str(round(self.alt[j]/ft)),
                                    str(round(self.cas[j]/kts))])
            # Savecmd(cmd,line): line is saved, cmd is used to prevent recording PAN & ZOOM commands and CRE
            # So insert a dummy command to record the line
            savecmd("---",line)

        # Check for crecmdlist: contains commands to be issued for this a/c
        # If any are there, then stack them for all aircraft
        for j in range(self.ntraf - n, self.ntraf):
            for cmdtxt in self.crecmdlist:
                 bs.stack.stack(self.id[j]+" "+cmdtxt)

        return True

    def creconfs(self, acid, actype, targetidx, dpsi, dcpa, tlosh, dH=None, tlosv=None, spd=None):
        ''' Create an aircraft in conflict with target aircraft.

            Arguments:
            - acid: callsign of new aircraft
            - actype: aircraft type of new aircraft
            - targetidx: id (callsign) of target aircraft
            - dpsi: Conflict angle (angle between tracks of ownship and intruder) (deg)
            - cpa: Predicted distance at closest point of approach (NM)
            - tlosh: Horizontal time to loss of separation ((hh:mm:)sec)
            - dH: Vertical distance (ft)
            - tlosv: Vertical time to loss of separation
            - spd: Speed of new aircraft (CAS/Mach, kts/-)
        '''
        latref  = self.lat[targetidx]  # deg
        lonref  = self.lon[targetidx]  # deg
        altref  = self.alt[targetidx]  # m
        trkref  = radians(self.trk[targetidx])
        gsref   = self.gs[targetidx]   # m/s
        tasref  = self.tas[targetidx]   # m/s
        vsref   = self.vs[targetidx]   # m/s
        cpa     = dcpa * nm
        pzr     = bs.settings.asas_pzr * nm
        pzh     = bs.settings.asas_pzh * ft
        trk     = trkref + radians(dpsi)

        if dH is None:
            acalt = altref
            acvs  = 0.0
        else:
            acalt = altref + dH
            tlosv = tlosh if tlosv is None else tlosv
            acvs  = vsref - np.sign(dH) * (abs(dH) - pzh) / tlosv

        if spd:
            # CAS or Mach provided: convert to groundspeed, assuming that
            # wind at intruder position is similar to wind at ownship position
            tas = tasref if spd is None else casormach2tas(spd, acalt)
            tasn, tase = tas * cos(trk), tas * sin(trk)
            wn, we = self.wind.getdata(latref, lonref, acalt)
            gsn, gse = tasn + wn, tase + we
        else:
            # Groundspeed is the same as ownship
            gsn, gse = gsref * cos(trk), gsref * sin(trk)

        # Horizontal relative velocity vector
        vreln, vrele = gsref * cos(trkref) - gsn, gsref * sin(trkref) - gse
        # Relative velocity magnitude
        vrel    = sqrt(vreln * vreln + vrele * vrele)
        # Relative travel distance to closest point of approach
        drelcpa = tlosh * vrel + (0 if cpa > pzr else sqrt(pzr * pzr - cpa * cpa))
        # Initial intruder distance
        dist    = sqrt(drelcpa * drelcpa + cpa * cpa)
        # Rotation matrix diagonal and cross elements for distance vector
        rd      = drelcpa / dist
        rx      = cpa / dist
        # Rotate relative velocity vector to obtain intruder bearing
        brn     = degrees(atan2(-rx * vreln + rd * vrele,
                                 rd * vreln + rx * vrele))

        # Calculate intruder lat/lon
        aclat, aclon = geo.kwikpos(latref, lonref, brn, dist / nm)
        # convert groundspeed to CAS, and track to heading using actual
        # intruder position
        wn, we     = self.wind.getdata(aclat, aclon, acalt)
        tasn, tase = gsn - wn, gse - we
        acspd      = tas2cas(sqrt(tasn * tasn + tase * tase), acalt)
        achdg      = degrees(atan2(tase, tasn))

        # Create and, when necessary, set vertical speed
        self.cre(acid, actype, aclat, aclon, achdg, acalt, acspd)
        self.ap.selaltcmd(len(self.lat) - 1, altref, acvs)
        self.vs[-1] = acvs

    def delete(self, idx):
        """Delete an aircraft"""
        # If this is a multiple delete, sort first for list delete
        # (which will use list in reverse order to avoid index confusion)
        if isinstance(idx, Collection):
            idx = np.sort(idx)

        # Call the actual delete function
        super().delete(idx)

        # Update number of aircraft
        self.ntraf = len(self.lat)
        return True

    def update(self):
        # Update only if there is traffic ---------------------
        if self.ntraf == 0:
            return

        #---------- Atmosphere --------------------------------
        self.p, self.rho, self.Temp = vatmos(self.alt)

        #---------- ADSB Update -------------------------------
        self.adsb.update()

        #---------- Fly the Aircraft --------------------------
        self.ap.update()  # Autopilot logic

        # Conflict detection and resolution
        if self.asastimer.readynext():
            self.cd.update(self, self)
            self.cr.update(self.cd, self, self)

        self.aporasas.update()   # Decide to use autopilot or ASAS for commands

        #---------- Limit commanded speeds based on performance ------------------------------
        self.aporasas.tas, self.aporasas.vs, self.aporasas.alt = \
            self.perf.limits(self.aporasas.tas, self.aporasas.vs,
                             self.aporasas.alt, self.ax)

        #---------- Kinematics --------------------------------
        self.update_airspeed()
        self.update_groundspeed()
        self.update_pos()

        #---------- Simulate Turbulence -----------------------
        self.turbulence.update()

        # Check whether new traffic state triggers conditional commands
        self.cond.update()

        #---------- Aftermath ---------------------------------
        self.trails.update()

    def update_airspeed(self):
        # Compute horizontal acceleration
        delta_spd = self.aporasas.tas - self.tas
        need_ax = np.abs(delta_spd) > np.abs(bs.sim.simdt * self.perf.axmax)
        self.ax = need_ax * np.sign(delta_spd) * self.perf.axmax
        # Update velocities
        self.tas = np.where(need_ax, self.tas + self.ax * bs.sim.simdt, self.aporasas.tas)
        self.cas = vtas2cas(self.tas, self.alt)
        self.M = vtas2mach(self.tas, self.alt)

        # Turning bank triangle
        # tan phi = a centrigugal/a grav = omega^2 * R / g = omega * V /g
        # => omega = (g tan phi)/V
        turnrate = np.degrees(g0 * np.tan(np.where(self.ap.turnphi>self.eps*self.eps,
                                                   self.ap.turnphi,self.ap.bankdef)) \
                                          / np.maximum(self.tas, self.eps))
        delhdg = (self.aporasas.hdg - self.hdg + 180) % 360 - 180  # [deg]
        self.swhdgsel = np.abs(delhdg) > np.abs(bs.sim.simdt * turnrate)

        # Update heading
        self.hdg = np.where(self.swhdgsel, 
                            self.hdg + bs.sim.simdt * turnrate * np.sign(delhdg), self.aporasas.hdg) % 360.0

        # Update vertical speed (alt select, capture and hold autopilot mode)
        delta_alt = self.aporasas.alt - self.alt
        # Old dead band version:
        #        self.swaltsel = np.abs(delta_alt) > np.maximum(
        #            10 * ft, np.abs(2 * bs.sim.simdt * self.vs))

        # Update version: time based engage of altitude capture (to adapt for UAV vs airliner scale)
        self.swaltsel = np.abs(delta_alt) >  1.05*np.maximum(np.abs(bs.sim.simdt * self.aporasas.vs), \
                                                         np.abs(bs.sim.simdt * self.vs))
        target_vs = self.swaltsel * np.sign(delta_alt) * np.abs(self.aporasas.vs)
        delta_vs = target_vs - self.vs
        # print(delta_vs / fpm)
        need_az = np.abs(delta_vs) > 300 * fpm   # small threshold
        self.az = need_az * np.sign(delta_vs) * (300 * fpm)   # fixed vertical acc approx 1.6 m/s^2
        self.vs = np.where(need_az, self.vs+self.az*bs.sim.simdt, target_vs)
        self.vs = np.where(np.isfinite(self.vs), self.vs, 0)    # fix vs nan issue

    def update_groundspeed(self):
        # Compute ground speed and track from heading, airspeed and wind
        if self.wind.winddim == 0:  # no wind
            self.gsnorth  = self.tas * np.cos(np.radians(self.hdg))
            self.gseast   = self.tas * np.sin(np.radians(self.hdg))

            self.gs  = self.tas
            self.trk = self.hdg
            self.windnorth[:], self.windeast[:] = 0.0,0.0

        else:
            applywind = self.alt>50.*ft # Only apply wind when airborne

            vnwnd,vewnd = self.wind.getdata(self.lat, self.lon, self.alt)
            self.windnorth[:], self.windeast[:] = vnwnd,vewnd
            self.gsnorth  = self.tas * np.cos(np.radians(self.hdg)) + self.windnorth*applywind
            self.gseast   = self.tas * np.sin(np.radians(self.hdg)) + self.windeast*applywind

            self.gs  = np.logical_not(applywind)*self.tas + \
                       applywind*np.sqrt(self.gsnorth**2 + self.gseast**2)

            self.trk = np.logical_not(applywind)*self.hdg + \
                       applywind*np.degrees(np.arctan2(self.gseast, self.gsnorth)) % 360.

        self.work += (self.perf.thrust * bs.sim.simdt * np.sqrt(self.gs * self.gs + self.vs * self.vs))


    def update_pos(self):
        # Update position
        self.alt = np.where(self.swaltsel, np.round(self.alt + self.vs * bs.sim.simdt, 6), self.aporasas.alt)
        self.lat = self.lat + np.degrees(bs.sim.simdt * self.gsnorth / Rearth)
        self.coslat = np.cos(np.deg2rad(self.lat))
        self.lon = self.lon + np.degrees(bs.sim.simdt * self.gseast / self.coslat / Rearth)
        self.distflown += self.gs * bs.sim.simdt

    def id2idx(self, acid):
        """Find index of aircraft id"""
        if not isinstance(acid, str):
            # id2idx is called for multiple id's
            # Fast way of finding indices of all ACID's in a given list
            tmp = dict((v, i) for i, v in enumerate(self.id))
            return [tmp.get(acidi, -1) for acidi in acid]
        else:
             # Catch last created id (* or # symbol)
            if acid in ('#', '*'):
                return self.ntraf - 1

            try:
                return self.id.index(acid.upper())
            except:
                return -1

    def setnoise(self, noise=None):
        """Noise (turbulence, ADBS-transmission noise, ADSB-truncated effect)"""
        if noise is None:
            return True, "Noise is currently " + ("on" if self.turbulence.active else "off")

        self.turbulence.setnoise(noise)
        self.adsb.setnoise(noise)
        return True

    def engchange(self, acid, engid):
        """Change of engines"""
        self.perf.engchange(acid, engid)
        return

    def move(self, idx, lat, lon, alt=None, hdg=None, casmach=None, vspd=None):
        self.lat[idx]      = lat
        self.lon[idx]      = lon

        if alt is not None:
            self.alt[idx]    = alt
            self.selalt[idx] = alt

        if hdg is not None:
            self.hdg[idx]  = hdg
            self.ap.trk[idx] = hdg

        if casmach is not None:
            self.tas[idx], self.selspd[idx], _ = vcasormach(casmach, alt)

        if vspd is not None:
            self.vs[idx]     = vspd
            self.swvnav[idx] = False

    def poscommand(self, idxorwp: int|str):
        """POS command: Show info or an aircraft, airport, waypoint or navaid"""
        if isinstance(idxorwp, str):
            # Argument is a waypoint: airport, navaid or fix
            wp = idxorwp.upper()
            lines = "Info on "+wp+":\n"

            # First try airports (most used and shorter, hence faster list)
            iap = bs.navdb.getaptidx(wp)
            if iap>=0:
                aptypes = ["large","medium","small"]
                lines = lines + bs.navdb.aptname[iap]+"\n"                 \
                        + "is a "+ aptypes[max(-1,bs.navdb.aptype[iap]-1)] \
                        +" airport at:\n"                                    \
                        + latlon2txt(bs.navdb.aptlat[iap],                 \
                                     bs.navdb.aptlon[iap]) + "\n"          \
                        + "Elevation: "                                      \
                        + str(int(round(bs.navdb.aptelev[iap]/ft)))        \
                        + " ft \n"

               # Show country name
                try:
                    ico = bs.navdb.cocode2.index(bs.navdb.aptco[iap].upper())
                    lines = lines + "in "+bs.navdb.coname[ico]+" ("+      \
                             bs.navdb.aptco[iap]+")"
                except:
                    ico = -1
                    lines = lines + "Country code: "+bs.navdb.aptco[iap]
                try:
                    runways = bs.navdb.rwythresholds[bs.navdb.aptid[iap]].keys()
                    if runways:
                        lines = lines + "\nRunways: " + ", ".join(runways)
                except KeyError:
                    pass

            # Not found as airport, try waypoints & navaids
            else:
                iwps = bs.navdb.getwpindices(wp,bs.ref.lat,bs.ref.lon)
                if iwps[0]>=0:
                    typetxt = ""
                    desctxt = ""
                    lastdesc = "XXXXXXXX"
                    for i in iwps:

                        # One line type text
                        if typetxt == "":
                            typetxt = typetxt+bs.navdb.wptype[i]
                        else:
                            typetxt = typetxt+" and "+bs.navdb.wptype[i]

                        # Description: multi-line
                        samedesc = bs.navdb.wpdesc[i]==lastdesc
                        if desctxt == "":
                            desctxt = desctxt +bs.navdb.wpdesc[i]
                            lastdesc = bs.navdb.wpdesc[i]
                        elif not samedesc:
                            desctxt = desctxt +"\n"+bs.navdb.wpdesc[i]
                            lastdesc = bs.navdb.wpdesc[i]

                        # Navaid: frequency
                        if bs.navdb.wptype[i] in ["VOR","DME","TACAN"] and not samedesc:
                            desctxt = desctxt + " "+ str(bs.navdb.wpfreq[i])+" MHz"
                        elif bs.navdb.wptype[i]=="NDB" and not samedesc:
                            desctxt = desctxt+ " " + str(bs.navdb.wpfreq[i])+" kHz"

                    iwp = iwps[0]

                    # Basic info
                    lines = lines + wp +" is a "+ typetxt       \
                           + " at\n"\
                           + latlon2txt(bs.navdb.wplat[iwp],  \
                                        bs.navdb.wplon[iwp])
                    # Navaids have description
                    if len(desctxt)>0:
                        lines = lines+ "\n" + desctxt

                    # VOR give variation
                    if bs.navdb.wptype[iwp]=="VOR":
                        lines = lines + "\nVariation: "+ \
                                     str(bs.navdb.wpvar[iwp])+" deg"


                    # How many others?
                    nother = bs.navdb.wpid.count(wp)-len(iwps)
                    if nother>0:
                        verb = ["is ","are "][min(1,max(0,nother-1))]
                        lines = lines +"\nThere "+verb + str(nother) +\
                                   " other waypoint(s) also named " + wp

                    # In which airways?
                    connect = bs.navdb.listconnections(wp, \
                                                bs.navdb.wplat[iwp],
                                                bs.navdb.wplon[iwp])
                    if len(connect)>0:
                        awset = set([])
                        for c in connect:
                            awset.add(c[0])

                        lines = lines+"\nAirways: "+"-".join(awset)


               # Try airway id
                else:  # airway
                    awid = wp
                    airway = bs.navdb.listairway(awid)
                    if len(airway)>0:
                        lines = ""
                        for segment in airway:
                            lines = lines+"Airway "+ awid + ": " + \
                                    " - ".join(segment)+"\n"
                        lines = lines[:-1] # cut off final newline
                    else:
                        return False,idxorwp+" not found as a/c, airport, navaid or waypoint"

        elif idxorwp >= 0:
            # Argument is an aircraft id
            idx           = idxorwp
            acid          = self.id[idx]
            actype        = self.type[idx]
            latlon        = latlon2txt(self.lat[idx], self.lon[idx])
            alt           = round(self.alt[idx] / ft)
            hdg           = round(self.hdg[idx])
            trk           = round(self.trk[idx])
            cas           = round(self.cas[idx] / kts)
            tas           = round(self.tas[idx] / kts)
            gs            = round(self.gs[idx]/kts)
            M             = self.M[idx]
            VS            = round(self.vs[idx]/ft*60.)
            route         = self.ap.route[idx]

            # Position report
            lines = "Info on %s %s index = %d\n" %(acid, actype, idx)     \
                  + "Pos: "+latlon+ "\n"                                  \
                  + "Hdg: %03d   Trk: %03d\n"        %(hdg, trk)              \
                  + "Alt: %d ft  V/S: %d fpm\n"  %(alt,VS)                \
                  + "CAS/TAS/GS: %d/%d/%d kts   M: %.3f\n"%(cas,tas,gs,M)

            # FMS AP modes
            if self.swlnav[idx] and route.nwp > 0 and route.iactwp >= 0:

                if self.swvnav[idx]:
                    if self.swvnavspd[idx]:
                        lines = lines + "VNAV (incl.VNAVSPD), "
                    else:
                        lines = lines + "VNAV (NOT VNAVSPD), "

                lines += "LNAV to " + route.wpname[route.iactwp] + "\n"

            # Flight info: Destination and origin
            if self.ap.orig[idx] != "" or self.ap.dest[idx] != "":
                lines = lines +  "Flying"

                if self.ap.orig[idx] != "":
                    lines = lines +  " from " + self.ap.orig[idx]

                if self.ap.dest[idx] != "":
                    lines = lines +  " to " + self.ap.dest[idx]

            # Show a/c info and highlight route of aircraft in radar window
            # and pan to a/c (to show route)
            bs.scr.showroute(acid)

        # Show what we found on aircraft, airport and navaid/waypoint
        return True, lines

    def airwaycmd(self, key):
        ''' Show conections of a waypoint or airway. '''
        if bs.navdb.awid.count(key) > 0:
            return self.poscommand(key)

        # Find connecting airway legs
        wpid = key
        iwp = bs.navdb.getwpidx(wpid, bs.ref.lat, bs.ref.lon)
        if iwp < 0:
            return False,key + " not found."

        wplat = bs.navdb.wplat[iwp]
        wplon = bs.navdb.wplon[iwp]
        connect = bs.navdb.listconnections(key, wplat, wplon)
        if connect:
            lines = ""
            for c in connect:
                if len(c) >= 2:
                    # Add airway, direction, waypoint
                    lines = lines + c[0] + ": to " + c[1] + "\n"
            return True, lines[:-1]  # exclude final newline
        return False, f"No airway legs found for {key}"

    def settrans(self, alt=-999.):
        """ Set or show transition level"""
        # in case a valid value is ginve set it
        if alt > -900.:
            if alt > 0.:
                self.translvl = alt
                return True
            return False,"Transition level needs to be ft/FL and larger than zero"

        # In case no value is given, show it
        tlvl = int(round(self.translvl/ft))
        return True, f"Transition level = {tlvl}/FL{int(round(tlvl/100.))}"

    def setbanklim(self, idx, bankangle=None):
        ''' Set bank limit for given aircraft. '''
        if bankangle:
            self.ap.bankdef[idx] = np.radians(bankangle) # [rad]
            return True
        return True, f"Banklimit of {self.id[idx]} is {int(np.degrees(self.ap.bankdef[idx]))} deg"

    def setthrottle(self,idx,throttle=""):
        """Set throttle to given value or AUTO, meaning autothrottle on (default)"""

        if throttle:
            if throttle in ('AUTO', 'OFF'): # throttle mode off, ATS on
                self.swats[idx] = True   # Autothrottle on
                self.thr[idx] = -999.    # Set to invalid

            elif throttle == "IDLE":
                self.swats[idx] = False
                self.thr[idx] = 0.0

            else:
                # Check for percent unit
                if throttle.count("%")==1:
                    throttle= throttle.replace("%","")
                    factor = 0.01
                else:
                    factor = 1.0

                # Remaining option is that it is a float, so try conversion
                try:
                    x = factor*float(throttle)
                except:
                    return False,"THR invalid argument "+throttle

                # Check whether value makes sense
                if x<0.0 or x>1.0:
                    return False, "THR invalid value " + throttle +". Needs to be [0.0 , 1.0]"

                 # Valid value, set throttle and disable autothrottle
                self.swats[idx] = False
                self.thr[idx] = x

            return True

        if self.swats[idx]:
            return True,"ATS of "+self.id[idx]+" is ON"
        return True, "ATS of " + self.id[idx] + " is OFF. THR is "+str(self.thr[idx])

    def crecmd(self,cmdline):
        """CRECMD command: list of commands to be issued for each aircraft after creation
           This commands adds a command to the list of default commands.
           """
        # Help text need or info on current list?
        if cmdline=="" or cmdline=="?":
            if len(self.crecmdlist)>0:
                allcmds = ""
                for i,txt in enumerate(self.crecmdlist):
                    if i==0:
                        allcmds = "[acid] "+txt
                    else:
                        allcmds +="; [acid] "+txt
                return True,"CRECMD list: "+allcmds
            else:
                return True,"CRECMD will add a/c specific commands to an aircraft after creation"
        # Command to be added to list
        else:
            self.crecmdlist.append(cmdline)
        return True, ""

    def clrcrecmd(self):
        """CRECMD command: list of commands to be issued for each aircraft after creation
           This commands adds a command to the list of default commands.
       """
        ncrecmd = len(self.crecmdlist)
        if ncrecmd==0:
            return True,"CLRCRECMD deletes all commands on clears command"
        else:
            self.crecmdlist = []
            return True,str("All",ncrecmd,"crecmd commands deleted.")



"""
Miscellaneous modules

Modules:
     txt2alt(txt): read altitude[ft] from txt (FL ot ft)
     txt2tas(spd,h): read CAS or Mach and convert to TAS for given altitude
     tim2txt(t)  : convert time[s] to HH:MM:SS.hh
     i2txt(i,n)  : convert integer to string of n chars with leading zeros

Created by  : Jacco M. Hoekstra
"""
from time import strftime, gmtime
import numpy as np

from .aero import cas2tas, mach2tas, kts, fpm, ft
from .geo import magdec


def txt2alt(txt):
    """Convert text to altitude in meter: also FL300 => 30000. as float"""
    # First check for FL otherwise feet
    try:
        if txt.upper()[:2] == 'FL' and len(txt) >= 4:  # Syntax check Flxxx or Flxx
            return 100.0 * int(txt[2:]) * ft
        return float(txt) * ft
    except ValueError:
        pass
    raise ValueError(f'Could not parse "{txt}" as altitude"')


def tim2txt(t):
    """Convert time to timestring: HH:MM:SS.hh"""
    return strftime("%H:%M:%S.", gmtime(t)) + i2txt(int((t - int(t)) * 100.), 2)


def txt2tim(txt):
    """Convert text to time in seconds:
       SS.hh
       MM:SS.hh
       HH.MM.SS.hh
    """
    timlst = txt.strip().split(":")

    try:
        # Always SS.hh
        t = float(timlst[-1])

        # MM
        if len(timlst) > 1 and timlst[-2]:
            t += 60.0 * int(timlst[-2])

        # HH
        if len(timlst) > 2 and timlst[-3]:
            t += 3600.0 * int(timlst[-3])

        return t
    except (ValueError, IndexError):
        raise ValueError(f'Could not parse "{txt}" as time')


def txt2bool(txt):
    ''' Convert string to boolean. '''
    ltxt = txt.lower()
    if ltxt in ('true', 'yes', 'y', '1', 'on'):
        return True
    if ltxt in ('false', 'no', 'n', '0', 'off'):
        return False
    raise ValueError(f'Could not parse {txt} as bool.')


def i2txt(i, n):
    """Convert integer to string with leading zeros to make it n chars long"""
    return f'{i:0{n}d}'


def txt2hdg(txt, lat=None, lon=None):
    ''' Convert text to true or magnetic heading.
    Modified by : Yaofu Zhou'''
    heading = float(txt.upper().replace("T", "").replace("M", ""))

    if "M" in txt.upper():
        if None in (lat, lon):
            raise ValueError('txt2hdg needs a reference latitude and longitude '
                             'when a magnetic heading is parsed.')
        magnetic_declination = magdec(lat, lon)
        heading = (heading + magnetic_declination) % 360.0

    return heading


def txt2vs(txt):
    ''' Convert text to vertical speed.

        Arguments:
        - txt: text string representing vertical speed in feet per minute.

        Returns:
        - Vertical Speed (float) in meters per second.
    '''
    return fpm * float(txt)


def txt2spd(txt):
    """ Convert text to speed, keep type (EAS/TAS/MACH) unchanged.

        Arguments:
        - txt: text string representing speed

        Returns:
        - Speed in meters per second or Mach.
    """
    try:
        txt = txt.upper()
        spd = float(txt.replace("M0.", ".").replace("M", ".").replace("..", "."))

        if not (0.1 < spd < 1.0 or txt.count("M") > 0):
            spd *= kts
        return spd
    except ValueError:
        raise ValueError(f'Could not parse {txt} as speed.')


def txt2tas(txt, h):
    """Convert text to speed (EAS [kts]/MACH[-] to TAS[m/s])"""
    if len(txt) == 0:
        return -1.
    try:
        if txt[0] == 'M':
            M_ = float(txt[1:])
            if M_ >= 20:   # Handle M95 notation as .95
                M_ = M_ * 0.01
            acspd = mach2tas(M_, h)  # m/s

        elif txt[0] == '.' or (len(txt) >= 2 and txt[:2] == '0.'):
            spd_ = float(txt)
            acspd = mach2tas(spd_, h)  # m/s

        else:
            spd_ = float(txt) * kts
            acspd = cas2tas(spd_, h)  # m/s
    except ValueError:
        return -1.

    return acspd


def col2rgb(txt):
    ''' Convert named color to R,G,B values (integer per component, 0-255) '''
    cols = {"black": (0, 0, 0), "white": (255, 255, 255), "green": (0, 255, 0),
            "red": (255, 0, 0), "blue": (0, 0, 255), "magenta": (255, 0, 255),
            "yellow": (240, 255, 127), "amber": (255, 255, 0), "cyan": (0, 255, 255)}
    try:
        rgb = cols[txt.lower().strip()]
    except KeyError:
        rgb = cols["white"]  # default

    return rgb

def degto180(angle):
    """Change to domain -180,180 degrees"""
    return (angle + 180.) % 360 - 180.

def radtopi(angle):
    """Change to domain -pi,pi radians """
    return (angle + np.pi) % (2.0 * np.pi) - np.pi

def findnearest(lat, lon, latarr, lonarr):
    """Find index of nearest postion in numpy arrays with lat and lon"""
    if len(latarr) > 0 and len(latarr) == len(lonarr):
        coslat = np.cos(np.radians(lat))
        dy = np.radians(lat - latarr)
        dx = coslat * np.radians(degto180(lon - lonarr))
        d2 = dx * dx + dy * dy
        idx = list(d2).index(d2.min())

        return idx
    return -1


def cmdsplit(cmdline, trafids=None):
    cmdline = cmdline.strip()
    if len(cmdline) == 0:
        return '', []

    # Use both comma and space as a separator: two commas mean an empty argument
    while cmdline.find(",,") >= 0:
        cmdline = cmdline.replace(",,", ",@,")  # Mark empty arguments

    # Replace comma's by space
    cmdline = cmdline.replace(",", " ")

    # Split using spaces
    cmdargs = cmdline.split()  # Make list of cmd arguments

    # Adjust for empty arguments
    for i in range(len(cmdargs)):
        if cmdargs[i] == "@":
            cmdargs[i] = ""

    # If a traffic id list is passed, check if command and first argument need to be switched
    if trafids and len(cmdargs) > 1 and trafids.count(cmdargs[0]):
        cmdargs[0:2] = cmdargs[1::-1]

    # return command, argumentlist
    return cmdargs[0], cmdargs[1:]


def txt2lat(lattxt):
    """txt2lat: input txt: N52'14'13.5 or N52 or N52' """
    txt = lattxt.upper().replace("N", "").replace("S", "-")  # North positive, South negative
    neg = txt.count("-") > 0

    # Use of "'" and '"' as delimiter for degrees/minutes/seconds
    # (also accept degree symbol chr(176))
    if txt.count("'") > 0 or txt.count('"') > 0 or txt.count(chr(176)) > 0:
        txt = txt.replace('"', "'").replace(chr(176),"'")# replace " or degree symbol and  by a '
        degs = txt.split("'")
        div = 1
        lat = 0
        if neg:
            f = -1.
        else:
            f = 1.
        for xtxt in degs:
            if len(xtxt) > 0:
                try:
                    lat = lat + f * abs(float(xtxt)) / float(div)
                    div = div * 60
                except ValueError:
                    print("txt2lat value error:",lattxt)
                    return 0.0
    else:
        lat = float(txt)
    return lat
    # Return float


def txt2lon(lontxt):
    """txt2lat: input txt: N52'14'13.5 or N52"""
    # It should first be checked if lontxt is a regular float, to avoid removing
    # the 'e' in a scientific-notation number.
    try:
        lon = float(lontxt)

    # Leading E will trigger error ansd means simply East,just as  W = West = Negative
    except ValueError:

        txt = lontxt.upper().replace("E", "").replace("W", "-")  # East positive, West negative
        neg = txt.count("-") > 0

        # Use of "'" and '"' as delimiter for degrees/minutes/seconds
        # (also accept degree symbol chr(176)). Also "W002'"
        if txt.count("'") > 0 or txt.count('"') or txt.count(chr(176))> 0:
            # replace " or degree symbol and  by a '
            txt = txt.replace('"', "'").replace(chr(176),"'")
            degs = txt.split("'")
            div = 1
            lon = 0.0
            if neg:
                f = -1.
            else:
                f = 1.
            for xtxt in degs:
                if len(xtxt)>0.0:
                    try:
                        lon = lon + f * abs(float(xtxt)) / float(div)
                    except ValueError:
                        print("txt2lon value error:",lontxt)
                        return 0.0

                div = div * 60
        else:  # Cope with "W65"without "'" or '"', also "-65" or "--65"
            try:
                neg = txt.count("-") > 0
                if neg:
                    f = -1.
                else:
                    f = 1.
                lon = f*abs(float(txt))
            except ValueError:
                print("txt2lon value error:",lontxt)
                return 0.0

    return lon

def lat2txt(lat):
    ''' Convert latitude into string (N/Sdegrees'minutes'seconds). '''
    d,m,s = float2degminsec(abs(lat))
    return "NS"[lat<0] + "%02d'%02d'"%(int(d),int(m))+str(s)+'"'

def lon2txt(lon):
    ''' Convert longitude into string (E/Wdegrees'minutes'seconds). '''
    d,m,s = float2degminsec(abs(lon))
    return "EW"[lon<0] + "%03d'%02d'"%(int(d),int(m))+str(s)+'"'

def latlon2txt(lat,lon):
    ''' Convert latitude and longitude in latlon string. '''
    return lat2txt(lat)+"  "+lon2txt(lon)

def deg180(dangle):
    """ Convert any difference in angles to interval [ -180,180 ) """
    return (dangle + 180.) % 360. - 180.

def float2degminsec(x):
    ''' Convert an angle into a string describing the angle in degrees,
        minutes, and seconds. '''
    deg     = int(x)
    minutes = int(x*60.) - deg *60.
    sec     = int(x*3600.) - deg*3600. - minutes*60.
    return deg,minutes,sec

def findall(lst,x):
    ''' Find indices of multiple occurences of x in lst. '''
    idx = []
    i = 0
    found = True
    while i<len(lst) and found:
        try:
            i = lst[i:].index(x)+i
            idx.append(i)
            i = i + 1
            found = True
        except ValueError:
            found = False
    return idx



""" QTGL Gui for BlueSky."""
import os
import sys

from PyQt6.QtCore import QTimer, qInstallMessageHandler, QT_VERSION_STR

from PyQt6.QtCore import QtMsgType
from PyQt6.QtWidgets import QApplication
from PyQt6.QtGui import QFont

import bluesky as bs
from bluesky.network.client import Client
from bluesky.ui.qtgl.mainwindow import MainWindow, Splash, DiscoveryDialog

bs.settings.set_variable_defaults(qt_verbosity=1)

print(('Using Qt ' + QT_VERSION_STR + ' for windows and widgets'))

def gui_msg_handler(msgtype, context, msg):
    if msgtype == QtMsgType.QtDebugMsg and bs.settings.qt_verbosity > 3:
        print('Qt debug message:', msg)
    elif msgtype == QtMsgType.QtInfoMsg and bs.settings.qt_verbosity > 2:
        print('Qt information message:', msg)
    elif msgtype == QtMsgType.QtWarningMsg and bs.settings.qt_verbosity > 1:
        print('Qt gui warning:', msg)
    elif msgtype == QtMsgType.QtCriticalMsg and bs.settings.qt_verbosity > 0:
        print('Qt gui critical error:', msg)
    elif msgtype == QtMsgType.QtFatalMsg:
        print('Qt gui fatal error:', msg)
    


def start(hostname=None):
    # Install message handler for Qt messages
    qInstallMessageHandler(gui_msg_handler)

    # Avoid Window hidpi scaling of fonts affecting view of commands
    os.environ["QT_ENABLE_HIGHDPI_SCALING"] = "1"
    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    #QApplication.setHighDpiScaleFactorRoundingPolicy(Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)

    # Start the Qt main object
    app = QApplication(sys.argv[:1])

    # Explicitly set font to avoid font loading warning dialogs
    app.setFont(QFont('Sans'))

    # Start the bluesky network client
    client = Client()
    network_timer = QTimer()
    network_timer.timeout.connect(client.update)
    network_timer.start(20)

    splash = Splash()
    splash.show()

    splash.showMessage('Constructing main window')
    app.processEvents()
    win = MainWindow(bs.mode)
    win.show()
    splash.showMessage('Done!')
    app.processEvents()
    splash.finish(win)
    # If this instance of the gui is started in client-only mode, show
    # server selection dialog
    if bs.mode == 'client' and hostname is None:
        dialog = DiscoveryDialog(client.client_id, win)
        dialog.show()

    else:
        client.connect(hostname=hostname)

    # Start the Qt main loop
    # app.exec_()
    app.exec()


''' Loader functions for navigation data. '''
import pickle

from bluesky import settings
from bluesky.tools import cachefile
from .loadnavdata_txt import loadnavdata_txt, loadthresholds_txt


# Cache versions: increment these to the current date if the source data is updated
# or other reasons why the cache needs to be updated
navdb_version = 'v20170101'

## Default settings
settings.set_variable_defaults(navdata_path='navdata')


def load_navdata():
    ''' Load navigation database. '''
    with cachefile.openfile('navdata.p', navdb_version) as cache:
        try:
            wptdata       = cache.load()
            awydata       = cache.load()
            aptdata       = cache.load()
            firdata       = cache.load()
            codata        = cache.load()
            rwythresholds = cache.load()
        except (pickle.PickleError, cachefile.CacheError) as e:
            print(e.args[0])

            wptdata, aptdata, awydata, firdata, codata = loadnavdata_txt()
            rwythresholds = loadthresholds_txt()

            cache.dump(wptdata)
            cache.dump(awydata)
            cache.dump(aptdata)
            cache.dump(firdata)
            cache.dump(codata)
            cache.dump(rwythresholds)

    return wptdata, aptdata, awydata, firdata, codata, rwythresholds



''' Core classes and functions based on which all other classes in BlueSky are
    made.
'''
from bluesky.core.base import Base, reset, select_implementation
from bluesky.core.entity import Entity
from bluesky.core.trafficarrays import TrafficArrays, RegisterElementParameters
from bluesky.core.signal import Signal
from bluesky.core.simtime import Timer
from bluesky.core.timedfunction import timed_function



''' BlueSky OpenGL classes and functions. '''
import importlib
from os import path
from collections import namedtuple
from collections import OrderedDict
from collections.abc import Collection, MutableMapping
import ctypes
import numpy as np
from pathlib import Path

from PyQt6.QtCore import qCritical
from PyQt6.QtOpenGLWidgets import QOpenGLWidget
from PyQt6.QtOpenGL import (QOpenGLShader, QOpenGLShaderProgram,
                        QOpenGLVertexArrayObject, QOpenGLBuffer,
                        QOpenGLVersionProfile, QOpenGLTexture, QOpenGLVersionFunctionsFactory)
from PyQt6.QtGui import QSurfaceFormat, QOpenGLContext, QImage

import bluesky as bs
from bluesky.core import Entity
from bluesky.stack import command
from bluesky.ui.qtgl.dds import DDSTexture


# Register settings defaults
bs.settings.set_variable_defaults(gfx_path='graphics')

GLVariable = namedtuple('GLVariable', ['loc', 'size'])

opengl_versions = ((4, 1), (2, 1), (2, 0))

# Convenience object with all GL functions
gl = None
_glvar_sizes = dict()


def get_profile_settings():
    for version in opengl_versions:
        for profile in ('Core', 'Compatibility'):
            try:
                importlib.import_module(f'PyQt6.QtOpenGL', package=f'QOpenGLFunctions_{version[0]}_{version[1]}_{profile}')
                
                print(f'Found Qt-provided OpenGL functions for OpenGL {version} {profile}')
                return version, QSurfaceFormat.OpenGLContextProfile.CoreProfile if profile == 'Core' else QSurfaceFormat.OpenGLContextProfile.CompatibilityProfile
            except:
                continue
    return (4, 1), None


def init():
    ''' Startup initialisation of OpenGL. '''
    if gl is None:
        version, profile = get_profile_settings()
        # Initialise application-wide GL version
        fmt = QSurfaceFormat()
        fmt.setVersion(*version)
        # profile = QSurfaceFormat.CoreProfile if sys.platform == 'darwin' else QSurfaceFormat.CompatibilityProfile

        fmt.setProfile(profile or QSurfaceFormat.OpenGLContextProfile.CompatibilityProfile)
        QSurfaceFormat.setDefaultFormat(fmt)

        if profile is not None:
            # Use a dummy context to get GL functions
            glprofile = QOpenGLVersionProfile(fmt)
            ctx = QOpenGLContext()
            function_factory = QOpenGLVersionFunctionsFactory()
            globals()['gl'] = function_factory.get(glprofile, ctx)
            # Check and set OpenGL capabilities
            if not glprofile.hasProfiles():
                raise RuntimeError(
                    'No OpenGL version >= 3.3 support detected for this system!')
        else:
            # If profile was none, PyQt6 is not shipped with any OpenGL function modules. Use PyOpenGL instead
            print("Couldn't find OpenGL functions in Qt. Falling back to PyOpenGL")
            globals()['gl'] = importlib.import_module('OpenGL.GL')

        globals()['_glvar_sizes'] = {
            gl.GL_FLOAT: 1, gl.GL_FLOAT_VEC2: 2, gl.GL_FLOAT_VEC3: 3,
            gl.GL_FLOAT_VEC4: 4, gl.GL_FLOAT_MAT2: 4, gl.GL_FLOAT_MAT3: 9,
            gl.GL_FLOAT_MAT4: 16, gl.GL_FLOAT_MAT2x3: 6, gl.GL_FLOAT_MAT2x4: 8,
            gl.GL_FLOAT_MAT3x2: 6, gl.GL_FLOAT_MAT3x4: 12, gl.GL_FLOAT_MAT4x2: 8,
            gl.GL_FLOAT_MAT4x3: 12, gl.GL_INT: 1, gl.GL_INT_VEC2: 2, gl.GL_INT_VEC3: 3,
            gl.GL_INT_VEC4: 4, gl.GL_UNSIGNED_INT: 1, gl.GL_UNSIGNED_INT_VEC2: 2,
            gl.GL_UNSIGNED_INT_VEC3: 3, gl.GL_UNSIGNED_INT_VEC4: 4, gl.GL_DOUBLE: 1,
            gl.GL_DOUBLE_VEC2: 2, gl.GL_DOUBLE_VEC3: 3, gl.GL_DOUBLE_VEC4: 4,
            gl.GL_DOUBLE_MAT2: 4, gl.GL_DOUBLE_MAT3: 9, gl.GL_DOUBLE_MAT4: 16,
            gl.GL_DOUBLE_MAT2x3: 6, gl.GL_DOUBLE_MAT2x4: 8, gl.GL_DOUBLE_MAT3x2: 6,
            gl.GL_DOUBLE_MAT3x4: 12, gl.GL_DOUBLE_MAT4x2: 8, gl.GL_DOUBLE_MAT4x3: 12}

    return gl


if gl is None:
    init()


def init_glcontext(ctx):
    ''' Correct OpenGL functions can only be obtained from a valid GL context. '''
    if getattr(gl, '__name__', '') != 'OpenGL.GL':
        # The OpenGL functions are provided by the Qt library. Update them from the current context
        fmt = QSurfaceFormat.defaultFormat()
        glprofile = QOpenGLVersionProfile(fmt)
        function_factory = QOpenGLVersionFunctionsFactory()
        globals()['gl'] = function_factory.get(glprofile, ctx)
    # QtOpenGL doesn't wrap all necessary functions. We can do this manually

    # void glGetActiveUniformBlockName(	GLuint program,
    #                                   GLuint uniformBlockIndex,
    #                                   GLsizei bufSize,
    #                                   GLsizei * length,
    #                                   GLchar * uniformBlockName)

    funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32,
                               ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_char * 20))
    funcptr = ctx.getProcAddress(b'glGetActiveUniformBlockName')
    c_getuboname = funtype(funcptr.__int__())

    def p_getuboname(programid, uboid):
        name = (ctypes.c_char * 20)()
        c_getuboname(programid, uboid, 20, None, ctypes.pointer(name))
        return name.value.decode('utf-8')

    gl.glGetActiveUniformBlockName = p_getuboname

    # void glGetActiveUniformBlockiv( GLuint program,
    #                                 GLuint uniformBlockIndex,
    #                                 GLenum pname,
    #                                 GLint * params)
    funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32,
                               ctypes.POINTER(ctypes.c_int32))
    funcptr = ctx.getProcAddress(b'glGetActiveUniformBlockiv')
    c_getuboiv = funtype(funcptr.__int__())

    def p_getuboiv(programid, uboid, pname):
        if pname == gl.GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES:
            # Special case: array with specific size is expected
            n_indices = ctypes.c_int32()
            c_getuboiv(programid, uboid,
                       gl.GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS, ctypes.pointer(n_indices))
            indices = (ctypes.c_int32 * n_indices.value)()
            # Apparently we need to specifically wrap depending on array size
            funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32,
                                       ctypes.POINTER(ctypes.c_int32 * n_indices.value))
            funcptr = ctx.getProcAddress(b'glGetActiveUniformBlockiv')
            c_getuboindices = funtype(funcptr.__int__())
            c_getuboindices(programid, uboid, pname, ctypes.pointer(indices))
            return list(indices)

        param = ctypes.c_int32()
        c_getuboiv(programid, uboid, pname, ctypes.pointer(param))
        return param.value

    gl.glGetActiveUniformBlockiv = p_getuboiv

    # void glVertexAttrib4Nub( GLuint index,
    #                          GLubyte v0,
    #                          GLubyte v1,
    #                          GLubyte v2,
    #                          GLubyte v3)
    funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_uint8,
                               ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
    funcptr = ctx.getProcAddress(b'glVertexAttrib4Nub')
    gl.glVertexAttrib4Nub = funtype(funcptr.__int__())

    # void glTexImage2D( GLenum target,
    #                    GLint level,
    #                    GLint internalFormat,
    #                    GLsizei width,
    #                    GLsizei height,
    #                    GLint border,
    #                    GLenum format,
    #                    GLenum type,
    #                    const void * data)
    funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_int32, ctypes.c_int32,
                               ctypes.c_uint32, ctypes.c_uint32, ctypes.c_int32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_void_p)
    funcptr = ctx.getProcAddress(b'glTexImage2D')
    gl.glTexImage2D_alt = funtype(funcptr.__int__())

    # void glTexSubImage2D(	GLenum target,
    #                       GLint level,
    #                       GLint xoffset,
    #                       GLint yoffset,
    #                       GLsizei width,
    #                       GLsizei height,
    #                       GLenum format,
    #                       GLenum type,
    #                       const void * pixels)
    funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_int32, ctypes.c_int32,
                               ctypes.c_int32, ctypes.c_uint32, ctypes.c_uint32,
                               ctypes.c_uint32, ctypes.c_uint32, ctypes.c_void_p)
    funcptr = ctx.getProcAddress(b'glTexSubImage2D')
    gl.glTexSubImage2D_alt = funtype(funcptr.__int__())
    
    if getattr(gl, '__name__', '') == 'OpenGL.GL':
        # In case we are using PyOpenGL, get some of the functions to behave in the
        # same way as their counterparts in Qt
        glGetProgramiv_wrap = gl.glGetProgramiv
        def glGetProgramiv(program, pname):
            params = ctypes.c_int32()
            glGetProgramiv_wrap(program, pname, params)
            return params.value
        gl.glGetProgramiv = glGetProgramiv

        glGetActiveAttrib_wrap = gl.glGetActiveAttrib
        def glGetActiveAttrib(program, index):
            length = ctypes.c_int32()
            size = ctypes.c_int32()
            atype = ctypes.c_uint32()
            name = (ctypes.c_char * 20)()
            glGetActiveAttrib_wrap(program, index, 20, length, size, atype, ctypes.pointer(name))
            return name.value.decode('utf-8'), size.value, atype.value
        gl.glGetActiveAttrib = glGetActiveAttrib

        glGetActiveUniform_wrap = gl.glGetActiveUniform
        def glGetActiveUniform(program, index):
            length = ctypes.c_int32()
            size = ctypes.c_int32()
            atype = ctypes.c_uint32()
            name = (ctypes.c_char * 20)()
            glGetActiveUniform_wrap(program, index, 20, length, size, atype, ctypes.pointer(name))
            return name.value.decode('utf-8'), size.value, atype.value
        gl.glGetActiveUniform = glGetActiveUniform

        glTexSubImage3D_wrap = gl.glTexSubImage3D
        def glTexSubImage3D(target, level, xOffset, yOffset, zOffset, width, height, depth, sourceFormat, sourceType, data):
            glTexSubImage3D_wrap(target, level, xOffset, yOffset, zOffset, width,
                                height, depth, sourceFormat, sourceType, ctypes.c_void_p(int(data)))
        gl.glTexSubImage3D = glTexSubImage3D

    
        # The extended initialisation of the remaining GL functions in this function
        # is only required for Qt-provided GL functions
        return

    # GLuint glGetUniformBlockIndex( GLuint program,
    #                                const GLchar * uniformBlockName)
    funtype = ctypes.CFUNCTYPE(
        ctypes.c_uint32, ctypes.c_uint32, ctypes.POINTER(ctypes.c_char))
    funcptr = ctx.getProcAddress(b'glGetUniformBlockIndex')
    c_getuboindex = funtype(funcptr.__int__())

    def p_getuboindex(programid, uboname):
        ret = c_getuboindex(programid, ctypes.c_char_p(
            uboname.encode('utf-8')))
        return ret

    gl.glGetUniformBlockIndex = p_getuboindex

    # void glVertexAttribIPointer(GLuint index,
    #                             GLint size,
    #                             GLenum type,
    #                             GLsizei stride,
    #                             const void * pointer)


    funtype = ctypes.CFUNCTYPE(None, ctypes.c_uint32, ctypes.c_int32, ctypes.c_uint32, 
                               ctypes.c_uint32, ctypes.c_void_p)
    funcptr = ctx.getProcAddress(b'glVertexAttribIPointer')
    # c_getvapointer = funtype(funcptr.__int__())
    # def p_getvapointer(index, size, atype, normalized, stride, offset):
    #     norm = ctypes.c_bool(0)
    #     c_getvapointer(index, size, atype, norm, stride, offset)
    gl.glVertexAttribIPointer = funtype(funcptr.__int__())


class ShaderSet(MutableMapping):
    ''' A set of shader programs for BlueSky.

        Convenience class to easily switch between sets of shader programs
        (e.g., between the radarwidget and the nd.)

        Normally, each set contains at least the following programs:
        'normal':   Rendering of regular, untextured shapes
        'textured': Rendering of regular, textured shapes
        'text':     Rendering of text objects
        'ssd':      Rendering of SSD objects
    '''
    # Currently selected shader set
    selected = None

    def __init__(self, parent=None):
        super().__init__()
        self.parent = parent
        self._programs = dict()
        self._ubos = dict()
        self._spath = ''
        self._iscreated = False
        if ShaderSet.selected is None:
            self.select()

    def create(self):
        ''' Overload this function for creation of shader programs in this set. '''
        self._iscreated = True

    def created(self):
        ''' Returns True if this shaderset was successfully created. '''
        return self._iscreated

    def select(self):
        ''' Select this shader set. '''
        ShaderSet.selected = self

    def update_ubo(self, uboname, *args, **kwargs):
        ''' Update an uniform buffer object of this shader set. '''
        ubo = self._ubos.get(uboname, None)
        if ubo is None:
            raise KeyError('Uniform Buffer Object', uboname,
                           'not found in shader set.')
        ubo.update(*args, **kwargs)
    def set_shader_path(self, shader_path):
        ''' Set a search path for shader files. '''
        self._spath = shader_path

    def load_shader(self, shader_name, vs, fs, gs=None):
        ''' Load a shader into this shader set.
            default shader names are: normal, textured, and text. '''
        vs = path.join(self._spath, vs)
        fs = path.join(self._spath, fs)
        if gs:
            gs = path.join(self._spath, gs)

        newshader = ShaderProgram(self.parent)
        if newshader.create(vs, fs, gs):
            self[shader_name] = newshader

    def release_all(self):
        ''' Release bound shader from context. '''
        if ShaderProgram.bound_shader:
            ShaderProgram.bound_shader.release()
            ShaderProgram.bound_shader = None

    def __getitem__(self, key):
        ret = self._programs.get(key, None)
        if not ret:
            raise KeyError('Shader program', key, 'not found in shader set.')
        return ret

    def __setitem__(self, key, program):
        if not isinstance(program, ShaderProgram):
            raise ValueError(
                'Only ShaderProgram objects can be added to a ShaderSet')
        self._programs[key] = program
        # Bind UBO buffers of this shader set to program's UBO's
        for name, size in program.ubos.items():
            ubo = self._ubos.get(name, None)
            if ubo is None:
                ubo = UniformBufferObject()
                ubo.create(size)
                self._ubos[name] = ubo

            program.bind_uniform_buffer(name, ubo)

    def __delitem__(self, key):
        del self._programs[key]

    def __iter__(self):
        return iter(self._programs)

    def __len__(self):
        return len(self._programs)

    @classmethod
    def get_shader(cls, shader_type=''):
        ''' Get a shader from the current shaderset by type. '''
        if not shader_type:
            return ShaderProgram.bound_shader
        ret = cls.selected._programs.get(shader_type, None)
        if not ret:
            raise KeyError('Shader program', shader_type,
                           'not found in shader set.')
        return ret


class ShaderProgram(QOpenGLShaderProgram):
    ''' BlueSky wrapper class for OpenGL shader programs. '''
    bound_shader = None

    def __init__(self, parent=None):
        super().__init__(parent)
        self.attribs = dict()
        self.ubos = dict()
        self.uniforms = dict()

    def bind(self):
        ''' Bind this shader to the current context. '''
        if ShaderProgram.bound_shader is not self:
            ShaderProgram.bound_shader = self
            super().bind()

    def release(self):
        ''' Release this shader from the current context. '''
        if ShaderProgram.bound_shader is self:
            super().release()
            ShaderProgram.bound_shader = None

    def create(self, fname_vertex, fname_frag, fname_geom=''):
        ''' Compile shaders and link program.
            Typically executed in initializeGL(). '''
        # Compile shaders from file
        success = True and \
            self.addShaderFromSourceFile(QOpenGLShader.ShaderTypeBit.Vertex, fname_vertex) and \
            self.addShaderFromSourceFile(QOpenGLShader.ShaderTypeBit.Fragment, fname_frag)
        if fname_geom:
            success = success and \
                self.addShaderFromSourceFile(QOpenGLShader.ShaderTypeBit.Geometry, fname_geom)

        # Link program
        if success:
            success = success and self.link()

        if not success:
            print('Shader program creation unsuccessful:')
            print(self.log())
            return False

        # Obtain list of attributes with location and size info
        n_attrs = gl.glGetProgramiv(self.programId(), gl.GL_ACTIVE_ATTRIBUTES)
        for a in range(n_attrs):
            name, size, atype = gl.glGetActiveAttrib(self.programId(), a)
            loc = self.attributeLocation(name)
            typesize = _glvar_sizes.get(atype, 1)
            self.attribs[name] = GLVariable(loc, size * typesize)

        # Get number of uniforms
        n_uniforms = gl.glGetProgramiv(self.programId(), gl.GL_ACTIVE_UNIFORMS)
        all_uids = set(range(n_uniforms))

        # Obtain list of uniform blocks
        n_ub = gl.glGetProgramiv(self.programId(), gl.GL_ACTIVE_UNIFORM_BLOCKS)
        for ub in range(n_ub):
            name = gl.glGetActiveUniformBlockName(self.programId(), ub)
            size = gl.glGetActiveUniformBlockiv(
                self.programId(), ub, gl.GL_UNIFORM_BLOCK_DATA_SIZE)
            self.ubos[name] = size
            # ubsize = gl.glGetActiveUniformBlockiv(self.programId(), ub,
            # gl.GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS)
            indices = gl.glGetActiveUniformBlockiv(
                self.programId(), ub, gl.GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES)
            all_uids -= set(indices)
            # print('Uniform block: ', name.value.decode('utf8'),
            #   'size =', ubsize.value)
            # for i in indices:
            #     usize = ctypes.c_int32()
            #     utype = ctypes.c_int32()
            #     uname = (ctpes.c_char * 20)()
            #     gl.glGetActiveUniform(self.programId(), i, 20, None,
            #         ctypes.pointer(usize), ctypes.pointer(utype),
            #         ctypes.pointer(uname))
            #     print('block uniform', i, '=', uname.value.decode('utf8'),
            #           'size =', usize.value)
            #     print(gl.glGetActiveUniform(self.programId(), i))

        # Obtain list of remaining uniforms (those not part of a block)
        for u in all_uids:
            name, size, utype = gl.glGetActiveUniform(self.programId(), u)
            self.uniforms[name] = GLVariable(
                u, size * _glvar_sizes.get(utype, 1))

        return True

    def bind_uniform_buffer(self, ubo_name, ubo):
        ''' Bind a uniform buffer block to this shader. '''
        idx = gl.glGetUniformBlockIndex(self.programId(), ubo_name)
        gl.glUniformBlockBinding(self.programId(), idx, ubo.binding)


class VertexArrayObject(QOpenGLVertexArrayObject):
    ''' Wrapper around the OpenGL approach to drawing a single shape from one
        or more buffers. '''

    def __init__(self, primitive_type=gl.GL_LINE_STRIP, shader_type='normal', parent=None):
        super().__init__(parent)
        self.shader_type = shader_type
        self.texture = None
        self.primitive_type = primitive_type
        self.first_vertex = 0
        self.vertex_count = 0
        self.n_instances = 0
        self.max_instance_divisor = 0
        self.single_color = None
        self.single_scale = None

    def set_primitive_type(self, primitive_type):
        ''' Set the primitive type for this VAO. '''
        self.primitive_type = primitive_type

    def set_vertex_count(self, count):
        ''' Set the vertex count for this VAO. '''
        self.vertex_count = int(count)

    def set_first_vertex(self, vertex):
        ''' Set the first vertex for this VAO. '''
        self.first_vertex = vertex

    def create(self, texture=None, vertex_count=0, n_instances=0, **attribs):
        ''' Create the actual VAO, attach passed attribs, and potentially
            create new buffers. '''
        super().create()
        if texture is not None:
            if isinstance(texture, Texture):
                self.texture = texture
            else:
                self.texture = Texture()
                self.texture.load(texture)
            if self.shader_type == 'normal':
                self.shader_type = 'textured'
        self.vertex_count = vertex_count
        self.n_instances = n_instances

        # Update special attribute 'scale' if necessary
        scaleattrib = ShaderSet.get_shader(self.shader_type).attribs.get('scale')
        if scaleattrib is not None:
            self.single_scale = (scaleattrib.loc, 1.0)

        self.set_attribs(**attribs)

    def set_attribs(self, usage=QOpenGLBuffer.UsagePattern.StaticDraw, instance_divisor=0,
                    datatype=None, stride=0, offset=None, normalize=False,
                    **attribs):
        ''' Set attributes for this VAO. '''
        if not attribs:
            return
        self.bind()
        program = ShaderSet.get_shader(self.shader_type)
        for name, data in attribs.items():
            attrib = program.attribs.get(name, None)
            if not attrib:
                raise KeyError('Unknown attribute ' + name +
                               ' for shader type ' + self.shader_type)

            dtype = datatype or (gl.GL_UNSIGNED_BYTE if name == 'color' else gl.GL_FLOAT)
            if name == 'color':
                normalize = True

            if isinstance(data, QOpenGLBuffer):
                # A previously created GL buffer is passed
                buf = data
                buf.bind()
            elif isinstance(data, Collection):
                # Color attribute has a special condition for a single color
                if name == 'color' and np.size(data) in (3, 4):
                    # Store attrib location and color
                    self.single_color = np.append(attrib.loc, data)
                    if len(data) == 3:
                        # Add full alpha if none is given
                        self.single_color = np.append(self.single_color, 255)
                    continue
                # If the input is an array create a new GL buffer
                buf = GLBuffer()
                buf.create(usage=usage, data=data)
            elif isinstance(data, int):
                buf = GLBuffer()
                buf.create(usage=usage, size=data)

            # Special attribs: scale and vertex
            if name == 'vertex' and isinstance(data, Collection):
                self.vertex_count = np.size(data) // 2
            elif name == 'scale':
                if isinstance(data, float):
                    self.single_scale = (attrib.loc, data)
                    continue
                else:
                    self.single_scale = None

            # Bind the buffer to the indicated attribute for this VAO
            program.enableAttributeArray(attrib.loc)
            if dtype == gl.GL_UNSIGNED_BYTE and not normalize:
                gl.glVertexAttribIPointer(attrib.loc, attrib.size, dtype, stride, offset)
            else:
                program.setAttributeBuffer(attrib.loc, dtype, 0, attrib.size)

            # For instanced data, indicate per how many instances
            # we move a step in the buffer (1=per instance)
            if instance_divisor > 0:
                gl.glVertexAttribDivisor(attrib.loc, instance_divisor)
                self.max_instance_divisor = max(
                    instance_divisor, self.max_instance_divisor)

            # Keep a reference to the buffer of this attribute
            setattr(self, name, buf)

    def update(self, **attribs):
        ''' Update one or more buffers for this object. '''
        for name, data in attribs.items():
            attrib = getattr(self, name, None)
            if not isinstance(attrib, GLBuffer):
                raise KeyError('Unknown attribute ' + name)
            # Special attribs: color and vertex
            if name == 'vertex' and isinstance(data, Collection):
                self.vertex_count = np.size(data) // 2

            # Update the buffer of the attribute
            attrib.update(data)

    def bind(self):
        shader = ShaderSet.get_shader(self.shader_type)
        shader.bind()
        super().bind()

    def draw(self, primitive_type=None, first_vertex=None, vertex_count=None, n_instances=None):
        ''' Draw this VAO. '''
        if primitive_type is None:
            primitive_type = self.primitive_type

        if first_vertex is None:
            first_vertex = self.first_vertex

        if vertex_count is None:
            vertex_count = self.vertex_count

        if n_instances is None:
            n_instances = self.n_instances

        if vertex_count == 0:
            return
        self.bind()

        if self.single_color is not None:
            gl.glVertexAttrib4Nub(*self.single_color)
        elif self.texture:
            self.texture.bind(0)

        if self.single_scale is not None:
            ShaderSet.get_shader(self.shader_type).setAttributeValue(*self.single_scale)

        if n_instances > 0:
            gl.glDrawArraysInstanced(
                primitive_type, first_vertex, vertex_count, n_instances * self.max_instance_divisor)
        else:
            gl.glDrawArrays(primitive_type, first_vertex, vertex_count)


class Text(VertexArrayObject):
    ''' Convenience class for a text object. '''

    def __init__(self, charsize=16.0, blocksize=None, font=None):
        super().__init__(primitive_type=gl.GL_TRIANGLES, shader_type='text')
        self.font = font or Font.getdefault()
        self.blocksize = blocksize or (0, 0)
        self.charsize = charsize

    def create(self, text, lat=None, lon=None, color=None, vertex_offset=None, instanced=False):
        ''' Create the Text VAO. '''
        super().create()
        if not self.font.isCreated():
            self.font.create()
        w, h = self.charsize, self.charsize * self.font.char_ar
        x, y = vertex_offset or (0.0, 0.0)
        if instanced:
            vertices, texcoords = self.font.char(x, y, w, h)
        else:
            vertices, texcoords = [], []
            for i, c in enumerate(text):
                v, t = self.font.char(x + i * w, y, w, h, ord(c))
                vertices += v
                texcoords += t

        self.set_attribs(vertex=np.array(vertices, dtype=np.float32),
                         texcoords=np.array(texcoords, dtype=np.float32))

        if instanced:
            self.set_attribs(texdepth=text, instance_divisor=1,
                             datatype=gl.GL_UNSIGNED_BYTE)
        divisor = self.blocksize[0] * self.blocksize[1] if instanced else 0
        if lat is not None and lon is not None:
            self.set_attribs(lat=lat, lon=lon, instance_divisor=divisor)

        if color:
            self.set_attribs(color=color, instance_divisor=divisor)

    def draw(self, first_vertex=None, vertex_count=None, n_instances=None):
        ''' Draw this text VAO. '''
        ShaderSet.get_shader(self.shader_type).bind()
        self.font.bind()
        self.font.set_char_size(self.charsize)
        if self.blocksize:
            self.font.set_block_size(self.blocksize)
        super().draw(first_vertex=first_vertex,
                     vertex_count=vertex_count, n_instances=n_instances)


class RenderTarget:
    ''' Wrapper class for OpenGL render targets, which can be widgets inside a bigger window,
        or independent GL windows. '''
    __rendertargets__ = dict()

    def __init__(self, shaderset=None):
        # Store all RenderTarget objects
        RenderTarget.__rendertargets__[self.__class__.__name__.upper()] = self
        self._renderobjs = OrderedDict()
        self._shaderset = shaderset

    def set_shaderset(self, shset):
        ''' Set the shaderset for this render target. '''
        self._shaderset = shset

    def addobject(self, obj):
        ''' Add a RenderObject to this target to draw.
            Objects are stored and drawn sorted by layer. '''
        self._renderobjs[obj.getbase().name()] = obj
        self._renderobjs = OrderedDict(sorted(self._renderobjs.items(),
                                             key=lambda o: o[1].layer))


class RenderWidget(QOpenGLWidget, RenderTarget):
    def initializeGL(self):
        init_glcontext(self.context())
        # First check for supported GL version
        gl_version = float(gl.glGetString(gl.GL_VERSION)[:3])
        if gl_version < 3.3:
            print(('OpenGL context created with GL version %.1f' % gl_version))
            qCritical("""Your system reports that it supports OpenGL up to version %.1f. The minimum requirement for BlueSky is OpenGL 3.3.
                Generally, AMD/ATI/nVidia cards from 2008 and newer support OpenGL 3.3, and Intel integrated graphics from the Haswell
                generation and newer. If you think your graphics system should be able to support GL>=3.3 please open an issue report
                on the BlueSky Github page (https://github.com/TUDelft-CNS-ATM/bluesky/issues)""" % gl_version)
            return

        if self._shaderset is None and self._renderobjs:
            qCritical("Cannot create render objects without an initialised shader set!")
            return

        # Initialise our shaderset if this hasn't been done yet
        if not self._shaderset.created():
            self._shaderset.create()

        # Call the create method of all registered objects
        for obj in self._renderobjs.values():
            obj.create()

    def paintGL(self):
        """Paint the scene."""
        # clear the framebuffer
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        # Draw our render objects in order of layer
        for obj in self._renderobjs.values():
            obj.draw()

        # Release shaders
        self._shaderset.release_all()


class RenderObject(Entity, skipbase=True):
    ''' Convenience singleton class for drawing different (nested) objects. '''
    # Known RenderObject base classes
    __renderobjs__ = dict()

    def __init_subclass__(cls, layer=0, skipbase=False):
        # All renderobjects are replaceable, but it is still possible to create an intermediate non-instantiable base class
        if not skipbase and not hasattr(cls, '_baseimpl'):
            # Store passed layer as a class variable
            cls.layer = layer
            cls.visible = True
            # Store all RenderObject base implementations
            RenderObject.__renderobjs__[cls.__name__.upper()] = cls

        return super().__init_subclass__(replaceable=True, skipbase=skipbase)

    def __init__(self, parent=None):
        super().__init__()
        self.parent = parent or self.getdefault().implinstance().parent
        self.glsurface = self.parent.glsurface if isinstance(self.parent, RenderObject) else self.parent
        self.children = list()

    def draw(self):
        ''' Draw this object. '''
        for child in self.children:
            child.draw()

    def create(self):
        ''' Create this object. '''
        pass

    @property
    def shaderset(self):
        ''' The shaderset of a RenderObject.
            Always points to the currently selected ShaderSet.
        '''
        return ShaderSet.selected


@command(aliases=('ADDVIS',))
def addvisual(objname: "txt" = "", target: "txt" = "RADARWIDGET"):
    ''' Add a new render object to a render target. 
        Use this function to enable the drawing of new GL render objects
        that are defined in plugins.

        Note that if you are reimplementing an existing GL object you
        should use the VIS command instead, to select alternate visualisations
        of existing GL objects.
    
        Arguments:
        - obj: The name of the renderobject to add. 
        - target: A render target such as the RadarWidget (the default) and the ND.
    '''
    if not target:
        return True, f'Available render targets: {", ".join(RenderTarget.__rendertargets__)}'
    
    targetobj = RenderTarget.__rendertargets__.get(target)
    if not targetobj:
        return False, f'Render target {target} not found!\n' + \
            f'Available render targets: {", ".join(RenderTarget.__rendertargets__)}'

    if not objname:
        existing = targetobj._renderobjs
        canadd = set(RenderObject.__renderobjs__.keys()) - set(existing)
        msg = f'Target {target} is currently drawing the following objects:\n'
        msg += ', '.join(existing) + '\n'
        if canadd:
            msg += f'Further objects that can be added to {target} are:\n'
            msg += ', '.join(canadd)
        else:
            msg += 'There are no further objects available to add.'
        return True, msg

    classobj = RenderObject.__renderobjs__.get(objname)
    if not classobj:
        return False, f'Unknown render object: {objname}!'
    # Check if object is already instantiated
    firsttime = not classobj.is_instantiated()
    obj = classobj(parent=targetobj)
    if firsttime:
        targetobj.makeCurrent()
        obj.create()
    targetobj.addobject(obj)


@command(aliases=('VIS',))
def visual(objname: "txt" = "", vis: "bool/txt" = ""):
    ''' Set the appearance and visibility of render objects. '''
    if not objname:
        return True, "Render objects in BlueSky:\n" + ", ".join(RenderObject.__renderobjs__)
    baseimpl = RenderObject.__renderobjs__.get(objname)
    if not baseimpl:
        return False, f"Render object {objname} not known!"
    # If vis is a boolean, this command is meant to toggle the visibility of the render object
    if isinstance(vis, bool):
        baseimpl.visible = vis
        return True, f'setting visibility for {objname} to {vis}'
    all_impls = baseimpl.derived()
    if vis == "":
        return True, f"{objname} has the following available implementations:\n" + ", ".join(all_impls)

    impl = all_impls.get(vis)
    if impl is None:
        return False, f'{vis} doesn\'t exist.\n' + \
            f"{objname} has the following available implementations:\n" + \
            ", ".join(all_impls)
    # Implementation exists, we can select it
    # Check first if implementation was selected before, otherwise create should be called
    # after construction
    firsttime = not impl.is_instantiated()
    impl.select()
    if firsttime:
        impl.instance().glsurface.makeCurrent()
        impl.instance().create()
    return True, f'Selected {vis} as visualisation for {objname}.'


class Circle(VertexArrayObject):
    ''' Convenience class for a circle. '''

    def __init__(self, shader_type='normal', parent=None):
        super().__init__(gl.GL_LINE_LOOP, shader_type, parent)

    def create(self, radius, nsegments=36, **attribs):
        ''' Create the Circle VAO.'''
        vertices = [(radius * np.cos(i / nsegments * 2.0 * np.pi),
                     radius * np.sin(i / nsegments * 2.0 * np.pi)) for i in range(nsegments)]
        # vcircle = np.array(np.transpose((
        #     radius * np.cos(np.linspace(0.0, 2.0 * np.pi, nsegments)),
        #     radius * np.sin(np.linspace(0.0, 2.0 * np.pi, nsegments)))), dtype=np.float32)
        super().create(vertex=np.array(vertices, dtype=np.float32), **attribs)

class Rectangle(VertexArrayObject):
    ''' Convenience class for a rectangle. '''

    def __init__(self, shader_type='normal', parent=None):
        super().__init__(gl.GL_LINE_LOOP, shader_type, parent)

    def create(self, w, h, fill=False):
        ''' Create the Rectangle VAO.'''
        if fill:
            self.set_primitive_type(gl.GL_TRIANGLE_FAN)
        vrect = np.array([(-0.5 * h, 0.5 * w), (-0.5 * h, -0.5 * w),
                          (0.5 * h, -0.5 * w), (0.5 * h, 0.5 * w)], dtype=np.float32)
        super().create(vertex=vrect)


class GLBuffer(QOpenGLBuffer):
    ''' Wrapper class for vertex and index buffers. '''

    def create(self, size=None, usage=QOpenGLBuffer.UsagePattern.StaticDraw, data=None):
        ''' Create the buffer. '''
        if size is None and data is None:
            raise ValueError(
                'Either a size or a set of data should be provided when creating a GL buffer')
        super().create()
        self.setUsagePattern(usage)
        self.bind()
        if data is not None:
            bufdata, size = content_and_size(data)
            self.allocate(bufdata, size)
        else:
            self.allocate(size)
        # Test if allocated size is as requested
        if self.size() != size:
            print(f'GLBuffer: Warning: could not allocate buffer of size {size}. Actual size is {self.size()}')

    def update(self, data, offset=0, size=None):
        ''' Send new data to this GL buffer. '''
        dbuf, dsize = content_and_size(data)
        size = size or dsize
        self.bind()
        if size > self.size() - offset:
            print(f'GLBuffer: Warning, trying to send more data ({size} bytes)'
                  f'to buffer than allocated size ({self.size()} bytes).')
            size = self.size() - offset
        
        self.write(offset, dbuf, size)
        # TODO: master branch has try/except for buffer writes after closing context


class UniformBufferObject(GLBuffer):
    ''' Wrapper class for uniform buffers. '''
    ufo_max_binding = 1

    def __init__(self):
        super().__init__(QOpenGLBuffer.Type.VertexBuffer)
        self.binding = 0

    def create(self, size=None, usage=QOpenGLBuffer.UsagePattern.StaticDraw, data=None):
        ''' Create this UBO. '''
        super().create(size, usage, data)
        self.binding = UniformBufferObject.ufo_max_binding
        UniformBufferObject.ufo_max_binding += 1
        gl.glBindBufferBase(gl.GL_UNIFORM_BUFFER,
                            self.binding, self.bufferId())


class Texture(QOpenGLTexture):
    ''' BlueSky OpenGL Texture class. '''
    def __init__(self, target=QOpenGLTexture.Target.Target2D):
        super().__init__(target)

    def load(self, fname):
        ''' Load the texture into GPU memory. '''
        fname = Path(fname)
        if fname.suffix.lower() == '.dds':
            tex = DDSTexture(fname)
            self.setFormat(QOpenGLTexture.TextureFormat.RGB_DXT1)
            self.setSize(tex.width, tex.height)
            self.setWrapMode(QOpenGLTexture.WrapMode.Repeat)
            self.allocateStorage()
            self.setCompressedData(len(tex.data), tex.data)
        else:
            self.setData(QImage(fname.as_posix()))
            self.setWrapMode(QOpenGLTexture.WrapMode.Repeat)

    def bind(self, unit=0):
        # Overload texture bind with default texture unit to make sure that, unless an explicit
        # texture unit is passed, the default is always zero.
        super().bind(unit)

    def setLayerData(self, layer, image):
        ''' For array textures, set the image data at given layer. '''
        ptr = image.constBits()
        ptr.setsize(image.width() * image.height() * 4)

        # xOffset, yOffset, zOffset, width, height, depth, sourceFormat, sourceType, data
        # self.setData(0, 0, i - 30, imgsize[0], imgsize[1], 1,
        #              QOpenGLTexture.BGRA, QOpenGLTexture.UInt8, ptr) #Qt >= 5.14
        gl.glTexSubImage3D(gl.GL_TEXTURE_2D_ARRAY, 0, 0, 0, layer,
                           image.width(), image.height(), 1, gl.GL_BGRA, gl.GL_UNSIGNED_BYTE, ptr)

class Font(Texture):
    ''' BlueSky class to implement a font using a GL Texture array. '''
    _fonts = list()

    def __init__(self):
        super().__init__(QOpenGLTexture.Target.Target2DArray)
        self.char_ar = 1.0
        self.loc_char_size = 0
        self.loc_block_size = 0
        Font._fonts.append(self)

    def create(self):
        ''' Create this font. '''
        txtshader = ShaderSet.get_shader('text')
        self.loc_char_size = txtshader.uniforms['char_size'].loc
        self.loc_block_size = txtshader.uniforms['block_size'].loc

        # Load the first image to get font size
        fname = (bs.resource(bs.settings.gfx_path) / 'font/{i}.png').as_posix()
        img = QImage(fname.format(i=32))
        imgsize = (img.width(), img.height())
        self.char_ar = float(imgsize[1]) / imgsize[0]

        super().create()
        self.setFormat(QOpenGLTexture.TextureFormat.RGBA8_UNorm)
        self.setSize(img.width(), img.height())
        self.setLayers(127 - 30)

        self.bind()
        self.allocateStorage()
        # gl.glTexImage3D(gl.GL_TEXTURE_2D_ARRAY, 0, gl.GL_RGBA8,
        #                 imgsize[0], imgsize[1], 127 - 30, 0, gl.GL_BGRA,
        #                 gl.GL_UNSIGNED_BYTE, None)
        self.setWrapMode(QOpenGLTexture.CoordinateDirection.DirectionS,
                         QOpenGLTexture.WrapMode.ClampToBorder)
        self.setWrapMode(QOpenGLTexture.CoordinateDirection.DirectionT,
                         QOpenGLTexture.WrapMode.ClampToBorder)
        self.setMinMagFilters(QOpenGLTexture.Filter.Linear, QOpenGLTexture.Filter.Linear)

        # We're using the ASCII range 32-126; space, uppercase, lower case,
        # numbers, brackets, punctuation marks
        for i in range(30, 127):
            img = QImage(fname.format(i=i)).convertToFormat(QImage.Format.Format_ARGB32)
            self.setLayerData(i - 30, img)

    @classmethod
    def getdefault(cls):
        ''' Get the default font. '''
        if not cls._fonts:
            return Font()
        return cls._fonts[0]

    def set_char_size(self, char_size):
        ''' Set the character size uniform. '''
        gl.glUniform2f(self.loc_char_size, char_size, char_size * self.char_ar)

    def set_block_size(self, block_size):
        ''' Set the block size uniform. '''
        gl.glUniform2i(self.loc_block_size, block_size[0], block_size[1])

    @staticmethod
    def char(x, y, w, h, c=32):
        ''' Convenience function to get vertices and texture coordinates for a
            single character. '''
        # Two triangles per character
        vertices = [(x, y + h), (x, y), (x + w, y + h),
                    (x + w, y + h), (x, y), (x + w, y)]
        texcoords = [(0, 0, c), (0, 1, c), (1, 0, c),
                     (1, 0, c), (0, 1, c), (1, 1, c)]
        return vertices, texcoords


def content_and_size(data):
    ''' Convenience function to get the correct variables to upload data to
        GL buffers. '''
    if isinstance(data, np.ndarray):
        return data, data.nbytes
    if isinstance(data, (ctypes.Structure, ctypes.Array)):
        # return ctypes.pointer(data), ctypes.sizeof(data)
        return bytes(data), ctypes.sizeof(data)
    return None, 0



""" 
Standard Dialog modules, using Tkinter

Created by  : Jacco M. Hoekstra
"""
import sys
try:
    from tkinter import *
    import tkinter.filedialog as filedialog
except:
    try:
        from Tkinter import *
        import Tkinter.tkFileDialog as filedialog
    except:
        from Tkinter import *
        import Tkinter.filedialog as filedialog
        
import os
import bluesky as bs

def fileopen():
    """returns filename of scenariofile selected"""
    cdir = os.getcwd()
    # load Tk crash on Mac
    # temporary work around mac without loading and file
    if sys.platform == 'darwin':
       return ''

    dirpath = bs.resource(bs.settings.scenario_path)
    os.chdir(dirpath)

    master = Tk()
    master.withdraw()  # hiding tkinter window
    master.focus_set()
    
    file_path = filedialog.askopenfilename(title="Open scenario file", 
        filetypes=[("Scenario files",".scn"),("All files",".*")],\
        initialdir=".")

    # Close Tk, return to working directory
    master.quit()
    os.chdir(cdir)
    if type(file_path)==str or type(file_path)==str:
        fpath = str(file_path)
    else:
        print("Warning: Unexpected type from FileDialog:",file_path)
        print(type(file_path))
        print("No file selected.")
        fpath = ""
#    print file_path,fpath
    return fpath



""" BlueSky plugin template. The text you put here will be visible
    in BlueSky as the description of your plugin. """
import numpy as np
# Import the global bluesky objects. Uncomment the ones you need
from bluesky import stack  #, settings, navdb, traf, sim, scr, tools
from bluesky import navdb
from bluesky.tools.aero import ft
from bluesky.tools import geo, areafilter

### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():

    # Addtional initilisation code

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'ILSGATE',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim',

        # Update interval in seconds. By default, your plugin's update function(s)
        # are called every timestep of the simulation. If your plugin needs less
        # frequent updates provide an update interval.
        'update_interval': 0.0,

        # The update function is called after traffic is updated. Use this if you
        # want to do things as a result of what happens in traffic. If you need to
        # something before traffic is updated please use preupdate.
        'update':          update,

        # If your plugin has a state, you will probably need a reset function to
        # clear the state in between simulations.
        'reset':         reset
        }

    stackfunctions = {
        # The command name for your function
        'ILSGATE': [
            # A short usage string. This will be printed if you type HELP <name> in the BlueSky console
            'ILSGATE Airport/runway',

            # A list of the argument types your function accepts. For a description of this, see ...
            'txt',

            # The name of your function in this plugin
            ilsgate,

            # a longer help text of your function.
            'Define an ILS approach area for a given runway.']
    }

    # init_plugin() should always return these two dicts.
    return config, stackfunctions


### Periodic update functions that are called by the simulation. You can replace
### this by anything, so long as you communicate this in init_plugin

def update():
    pass

def reset():
    pass

### Other functions of your plugin
def ilsgate(rwyname):
    if '/' not in rwyname:
        return False, 'Argument is not a runway ' + rwyname
    apt, rwy = rwyname.split('/RW')
    rwy = rwy.lstrip('Y')
    apt_thresholds = navdb.rwythresholds.get(apt)
    if not apt_thresholds:
        return False, 'Argument is not a runway (airport not found) ' + apt
    rwy_threshold = apt_thresholds.get(rwy)
    if not rwy_threshold:
        return False, 'Argument is not a runway (runway not found) ' + rwy
    # Extract runway threshold lat/lon, and runway heading
    lat, lon, hdg = rwy_threshold

    # The ILS gate is defined as a triangular area pointed away from the runway
    # First calculate the two far corners in cartesian coordinates
    cone_length = 50 # nautical miles
    cone_angle  = 20.0 # degrees
    lat1, lon1 = geo.qdrpos(lat, lon, hdg - 180.0 + cone_angle, cone_length)
    lat2, lon2 = geo.qdrpos(lat, lon, hdg - 180.0 - cone_angle, cone_length)
    coordinates = np.array([lat, lon, lat1, lon1, lat2, lon2])
    areafilter.defineArea('ILS' + rwyname, 'POLYALT', coordinates, top=4000*ft)



''' Load navigation data from text files.'''
import numpy as np
from zipfile import ZipFile
from bluesky import settings
import bluesky as bs
from bluesky.tools.aero import ft


REARTH_INV = 1.56961231e-7


def loadnavdata_txt():
    #----------  Read  nav.dat file (nav aids) ----------
    wptdata         = dict()
    wptdata['wpid']    = []              # identifier (string)
    wptdata['wplat']   = []              # latitude [deg]
    wptdata['wplon']   = []              # longitude [deg]
    wptdata['wptype']  = []              # Type "VOR","NDB","DME","TACAN" or "FIX"

    wptdata['wpelev']  = []              # elevation [m]
    wptdata['wpvar']   = []              # magnetic variation [deg]
    wptdata['wpfreq']  = []              # Navaid frequency kHz(NDB) or MHz(VOR)
    wptdata['wpdesc']  = []              # description


    with open(bs.resource(settings.navdata_path) / 'nav.dat', 'rb') as f:
        print("Reading nav.dat")

        for line in f:
            line = line.decode(encoding="ascii", errors="ignore").strip()
            # Skip empty lines or comments
            if len(line) == 0 or line[0] == "#":
                continue

            # Data line => Process fields of this record, separated by a comma
            # Example lines:
            # 2  58.61466599  125.42666626    451   522  30    0.0 A    Aldan NDB
            # 3  31.26894444 -085.72630556    334 11120  40   -3.0 OZR  CAIRNS VOR-DME
            # type    lat       lon           elev freq  ?     var id    desc
            #   0      1         2              3    4   5      6   7     8

            fields = line.split()

            # Valid line starst with integers
            if not fields[0].isdigit():
                continue # Next line

            # Get code for type of navaid
            itype = int(fields[0])

            # Type names
            wptypedict = {2:"NDB",3:"VOR",         \
                          4:"ILS",5:"LOC",6:"GS",  \
                          7:"OM",8:"MM",9:"IM",
                          12:"DME",13:"TACAN"}

           # Type code never larger than 20
            if itype not in list(wptypedict.keys()):
                continue # Next line

            wptype = wptypedict[itype]

            # Select types to read
            if wptype not in ["NDB","VOR","DME","TACAN"]:
                continue # Next line

            wptdata["wptype"].append(wptype)

            wptdata["wplat"].append(float(fields[1]))      # latitude [deg]
            wptdata["wplon"].append(float(fields[2]))      # longitude [deg]
            wptdata["wpelev"].append(float(fields[3])*ft)  # elevation [ft]

            if wptype=="NDB":
                wptdata["wpfreq"].append(int(fields[4]))   # NDB freq in kHz

            elif wptype in ["VOR","DME","TACAN"]:
                wptdata["wpfreq"].append(float(fields[4])/100.) # VOR freq in MHz
            else:
                wptdata["wpfreq"].append(0.0)

            if wptype in ["VOR","NDB"]:
                wptdata["wpvar"].append(float(fields[6])) # Magnetic variation in np.degrees
                wptdata["wpid"].append(fields[7]) # Id

            elif wptype in ["DME","TACAN"]:
                wptdata["wpvar"].append(0.0) # Magnetic variation not given
                wptdata["wpid"].append(fields[7]) # Id

            else:
                wptdata['wpvar'].append(0.0) # Magnetic variation in np.degrees
                wptdata['wpid'].append(" ")  # Id

            # Find description
            wpid = wptdata["wpid"][-1]
            try:
                idesc = line.index(wpid)+len(wpid)

                # Description at end of line may include spaces
                wptdata['wpdesc'].append(line[idesc:])  # Description
            except:
                wptdata['wpdesc'].append("   ")  # Description

    #----------  Read  fix.dat file ----------
    with open(bs.resource(settings.navdata_path) / 'fix.dat', 'rb') as f:
        print("Reading fix.dat")
        for line in f:
            line = line.decode(encoding="ascii", errors="ignore").strip()

            # Skip empty lines or comments
            if len(line) < 3 or line[0] == "#":
                continue

            # Start with valid 2 digit latitude -45. or 52.
            if not ((line[0]=="-" and line[3]==".") or line[2]==".")  :
                continue


            # Data line => Process fields of this record, separated by a comma
            # Example line:
            #  30.580372 -094.384169 FAREL
            fields = line.split()

            wptdata["wptype"].append("FIX")


            wptdata['wplat'].append(float(fields[0]))      # latitude [deg]
            wptdata['wplon'].append(float(fields[1]))      # longitude [deg]
            wptdata['wpid'].append(fields[2]) # Id

            # Not given for fixes but fill out tables for equal length
            wptdata['wpelev'].append(0.0)  # elevation [ft]
            wptdata['wpfreq'].append(0.0)  # Fix is no navaid, so no freq
            wptdata['wpvar'].append(0.0)   # Magnetic variation not given
            wptdata['wpdesc'].append("")  # Description

    # Convert lists for lat,lon to numpy-array for vectorised clipping
    wptdata['wplat']   = np.array(wptdata['wplat'])
    wptdata['wplon']   = np.array(wptdata['wplon'])

    #----------  Read  awy.dat file (airway legs) ----------
    awydata   = dict()

    awydata['awid']        = []              # airway identifier (string)
    awydata['awfromwpid']  = []              # from waypoint identifier (string)
    awydata['awfromlat']   = []              # from waypoint lat [deg](float)
    awydata['awfromlon']   = []              # from waypoint lon [deg](float)
    awydata['awtowpid']    = []              # to waypoint identifier (string)
    awydata['awtolat']     = []              # to waypoint lat [deg](float)
    awydata['awtolon']     = []              # to waypoint lon [deg](float)
    awydata['awndir']      = []              # number of directions (1 or 2)
    awydata['awlowfl']     = []              # lowest flight level (int)
    awydata['awupfl']      = []              # highest flight level (int)


    with open(bs.resource(settings.navdata_path) / 'awy.dat', 'rb') as f:
        print("Reading awy.dat")

        for line in f:
            line = line.decode(encoding="ascii", errors="ignore").strip()
            # Skip empty lines or comments
            if len(line) == 0 or line[0] == "#":
                continue

            fields = line.split()
            if len(fields) < 10:
                continue

            # Example line
            # ABAGO  56.291668  144.236667 GINOL  54.413334  142.011667 1 177 528 A218
            # fromfwp fromlat    fromlon    towp   tolat       tolon   ndir lowfl hghfl airwayid
            #   0        1          2         3      4           5     6     7     8       9

            # Second field should be float
            try:
                fromlat = float(fields[1])
            except:
                continue

            awydata['awfromwpid'].append(fields[0])         # from waypoint identifier (string)
            awydata['awfromlat'].append(fromlat)            # from latitude [deg]
            awydata['awfromlon'].append(float(fields[2]))   # from longitude [deg]

            awydata['awtowpid'].append(fields[3])           # to waypoint identifier (string)
            awydata['awtolat'].append(float(fields[4]))     # to latitude [deg]
            awydata['awtolon'].append(float(fields[5]))     # to longitude [deg]

            awydata['awndir'].append(int(fields[6]))        # number of directions (1 or 2)

            awydata['awlowfl'].append(int(fields[7]))       # number of directions (1 or 2)
            awydata['awupfl'].append(int(fields[8]))        # number of directions (1 or 2)

            if fields[9].find("-")<0:
                #only one airway uses this leg
                awydata['awid'].append(fields[9])
            else:
                # More airways use this leg => copy leg with all airway ids
                awids = fields[9].split("-")
                for i, awid in enumerate(awids):
                    awydata['awid'].append(awid.strip())
                    if i>0:
                        # Repeat last entry
                        for key in awydata:
                            if key!="awid":
                                awydata[key].append(awydata[key][-1])


        # Convert lat,lons to numpy arrays for easy clipping
        awydata['awfromlat'] = np.array(awydata['awfromlat'])
        awydata['awfromlon'] = np.array(awydata['awfromlon'])
        awydata['awtolat']   = np.array(awydata['awtolat'])
        awydata['awtolon']   = np.array(awydata['awtolon'])

    #----------  Read airports.dat file ----------
    aptdata           = dict()
    aptdata['apid']      = []              # 4 char identifier (string)
    aptdata['apname']    = []              # full name
    aptdata['aplat']     = []              # latitude [deg]
    aptdata['aplon']     = []              # longitude [deg]
    aptdata['apmaxrwy']  = []              # reference airport {string}
    aptdata['aptype']    = []              # type (int, 1=large, 2=medium, 3=small)
    aptdata['apco']      = []              # two char country code (string)
    aptdata['apelev']    = []              # field elevation ft-> m
    with open(bs.resource(settings.navdata_path) / 'airports.dat', 'rb') as f:
        types = {'L': 1, 'M': 2, 'S': 3}
        for line in f:
            line = line.decode(encoding="ascii", errors="ignore").strip()
            # Skip empty lines or comments
            if len(line) == 0 or line[0] == "#":
                continue

            # Data line => Process fields of this record, separated by a comma
            # Example line:
            # EHAM, SCHIPHOL, 52.309, 4.764, Large, 12467, NL
            #  [id]   [name] [lat]    [lon]  [type] [max rwy length in ft] [country code] [elevation]
            #   0        1     2        3       4          5                   6            7
            fields = line.split(",")

            # Skip airports without identifier in file and closed airports
            if fields[0].strip() == "" or fields[4].strip() == 'Closed':
                continue

            aptdata['apid'].append(fields[0].strip())  # id, no leading or trailing spaces
            aptdata['apname'].append(fields[1].strip())  # name, no leading or trailing spaces

            aptdata['aplat'].append(float(fields[2]))  # latitude [deg]
            aptdata['aplon'].append(float(fields[3]))  # longitude [deg]

            aptdata['aptype'].append(types[fields[4].strip()[0]])  # large=1, medium=2, small=3

            # Not all airports have rwy length (e.g. heliports)
            try:
                aptdata['apmaxrwy'].append(float(fields[5])*ft)  # max rwy ltgh [m]
            except:
                aptdata['apmaxrwy'].append(0.0)

            aptdata['apco'].append(fields[6].strip().lower()[:2])     # country code

            # Not all airports have elevation in data
            try:
                aptdata['apelev'].append(float(fields[7])*ft)  # apt elev [m]
            except:
                aptdata['apelev'].append(0.0)


    aptdata['aplat']    = np.array(aptdata['aplat'])
    aptdata['aplon']    = np.array(aptdata['aplon'])
    aptdata['apmaxrwy'] = np.array(aptdata['apmaxrwy'])
    aptdata['aptype']   = np.array(aptdata['aptype'])
    aptdata['apelev']   = np.array(aptdata['apelev'])

    #----------  Read FIR files ----------
    firdata         = dict()
    firdata['fir']     = []
    firdata['firlat0'] = []
    firdata['firlon0'] = []
    firdata['firlat1'] = []
    firdata['firlon1'] = []

    # Get fir names
    for filname in (bs.resource(settings.navdata_path) / 'fir').iterdir():
        if filname.suffix == '.txt':
            firname = filname.stem
            firdata['fir'].append([firname, [], []])
            with open(filname, 'rb') as f:
                for line in f:
                    rec = line.decode(encoding="ascii", errors="ignore").upper().strip()

                    if len(rec) == 0:
                        continue

                    latsign = 2 * int(line[0] == "N") - 1
                    latdeg  = float(line[1:4])
                    latmin  = float(line[5:7])
                    latsec  = float(line[8:14])
                    lat     = latsign*latdeg+latmin/60.+latsec/3600.

                    lonsign = 2 * int(line[15] == "E") - 1
                    londeg  = float(line[16:19])
                    lonmin  = float(line[20:22])
                    lonsec  = float(line[23:29])
                    lon     = lonsign*londeg+lonmin/60.+lonsec/3600.

                    # For drawing create a line from last lat,lon to current lat,lon
                    if len(firdata['fir'][-1][1]) > 0:  # skip first lat,lon
                        firdata['firlat0'].append(firdata['fir'][-1][1][-1])
                        firdata['firlon0'].append(firdata['fir'][-1][2][-1])
                        firdata['firlat1'].append(lat)
                        firdata['firlon1'].append(lon)

                    # Add to FIR record
                    firdata['fir'][-1][1].append(lat)
                    firdata['fir'][-1][2].append(lon)

    # Convert lat/lon lines to numpy arrays
    firdata['firlat0'] = np.array(firdata['firlat0'])
    firdata['firlat1'] = np.array(firdata['firlat1'])
    firdata['firlon0'] = np.array(firdata['firlon0'])
    firdata['firlon1'] = np.array(firdata['firlon1'])

    #----------  Read ICAO country codes file icao-countries.dat ----------
    codata           = dict()
    codata['coname']   = []              # Country name
    codata['cocode2']  = []              # 2 char code
    codata['cocode3']  = []              # 3 char code
    codata['conr']     = []              # country nr
    with open(bs.resource(settings.navdata_path) / 'icao-countries.dat', 'rb') as f:
        for line in f:
            line = line.decode(encoding="ascii", errors="ignore").strip()
            # Skip empty lines or comments
            if len(line) == 0 or line[0] == "#":
                continue

            # Data line: comma separated values:
            # full name, A2 code, A3 code, number

            fields = line.split(",")

            # Skip airports without identifier in file and closed airports
            if fields[0].strip() == "":
                continue

            codata['coname'].append(fields[0].strip())  # id, no leading or trailing spaces
            codata['cocode2'].append(fields[1].strip().upper())  # name, no leading or trailing spaces

            codata['cocode3'].append(fields[2].strip().upper())  # latitude [deg]
            try:
                codata['conr'].append(int(fields[3]))  # longitude [deg]
            except:
                codata['conr'].append(-1)


    return wptdata, aptdata, awydata, firdata, codata


def loadthresholds_txt():
    ''' Runway threshold loader for navdatabase. '''
    rwythresholds = dict()
    curthresholds = None
    zfile = ZipFile(bs.resource(settings.navdata_path) / 'apt.zip')
    print("Reading apt.dat from apt.zip")
    with zfile.open('apt.dat', 'r') as f:
        for line in f:
            elems = line.decode(
                encoding="ascii", errors="ignore").strip().split()
            if len(elems) == 0:
                continue

            # 1: AIRPORT
            if elems[0] == '1':
                # Add airport to runway threshold database
                curthresholds = dict()
                rwythresholds[elems[4]] = curthresholds
                continue

            if elems[0] == '100':
                # Only asphalt and concrete runways
                if int(elems[2]) > 2:
                    continue
                # rwy_lbl = (elems[8], elems[17])

                lat0 = float(elems[9])
                lon0 = float(elems[10])
                offset0 = float(elems[11])

                lat1 = float(elems[18])
                lon1 = float(elems[19])
                offset1 = float(elems[20])

                # threshold information: ICAO code airport, Runway identifier,
                # latitude, longitude, bearing
                # vertices: gives vertices of the box around the threshold

                # opposite runways are on the same line. RWY1: 8-11, RWY2: 17-20
                # Hence, there are two thresholds per line
                # thr0: First lat0 and lon0 , then lat1 and lat1, offset=[11]
                # thr1: First lat1 and lat1 , then lat0 and lon0, offset=[20]

                thr0 = thresholds(np.radians(lat0), np.radians(lon0),
                                  np.radians(lat1), np.radians(lon1), offset0)
                thr1 = thresholds(np.radians(lat1), np.radians(lon1),
                                  np.radians(lat0), np.radians(lon0), offset1)
                curthresholds[elems[8]] = thr0
                curthresholds[elems[17]] = thr1
                continue
    return rwythresholds


def thresholds(lat1, lon1, lat2, lon2, offset):
    ''' calculates the threshold points per runway
        underlying equations can be found at
        http://www.movable-type.co.uk/scripts/latlong.html '''

    d = offset * REARTH_INV
    deltal = lon2 - lon1

    # calculate runway bearing
    bearing = np.arctan2(np.sin(deltal) * np.cos(lat2), (np.cos(lat1) * np.sin(lat2) -
                                              np.sin(lat1) * np.cos(lat2) * np.cos(deltal)))

    # normalize to 0-360 degrees
    bearing = np.radians((np.degrees(bearing) + 360) % 360)

    # get threshold points
    latthres, lonthres = thrpoints(lat1, lon1, d, bearing)

    return np.degrees(latthres), np.degrees(lonthres), np.degrees(bearing)


def thrpoints(lat1, lon1, d, bearing):
    ''' Calculate threshold points as well as end points of threshold box
    underlying equations can be found at
    http://www.movable-type.co.uk/scripts/latlong.html '''
    latthres = np.arcsin(np.sin(lat1) * np.cos(d) + np.cos(lat1) * np.sin(d) * np.cos(bearing))

    lonthres = lon1 + np.arctan2(np.sin(bearing) * np.sin(d) * np.cos(lat1),
                            np.cos(d) - np.sin(lat1) * np.sin(latthres))

    return latthres, lonthres



''' BlueSky variable explorer

    Provide flexible access to simulation data in BlueSky.
'''
from numbers import Number
from collections import OrderedDict
from collections.abc import Collection
import re
import numpy as np
import bluesky as bs
from bluesky.core import TrafficArrays
from bluesky.core.entity import getproxied

# Globals
# The variable lists and their corresponding sources
varlist = OrderedDict()


def init():
    ''' Variable explorer initialization function.
        Is called in bluesky.init() '''
    # Add the default sources to the variable explorer
    varlist.update([('sim', (bs.sim, getvarsfromobj(bs.sim))),
                    ('traf', (bs.traf, getvarsfromobj(bs.traf)))])


def register_data_parent(obj, name):
    varlist[name] = (obj, getvarsfromobj(obj))


def getvarsfromobj(obj):
    ''' Return a list with the names of the variables of the passed object.'''
    try:
        # Return attribute names, but exclude private attributes
        return [name for name in vars(obj) if not name[0] == '_']
    except TypeError:
        return None


def lsvar(varname=''):
    ''' Stack function to list information on simulation variables in the
        BlueSky console. '''
    if not varname:
        # When no argument is passed, show a list of parent objects for which
        # variables can be accessed
        return True, '\n' + \
            str.join(', ', [key for key in varlist])

    # Find the variable in the variable list
    v = findvar(varname)
    if v:
        thevar = v.get()  # reference to the actual variable
        # When the variable is an object, get child attributes
        attrs = getvarsfromobj(thevar)
        vartype = v.get_type()  # Type of the variable
        if isinstance(v.parent, TrafficArrays) and v.parent.istrafarray(v.varname):
            vartype += ' (TrafficArray)'
        txt = \
            f'Variable:   {v.varname}\n' + \
            f'Type:       {vartype}\n'
        if isinstance(thevar, Collection):
            txt += f'Size:       {len(thevar)}\n'
        txt += f'Parent:     {v.parentname}'
        if attrs:
            txt += '\nAttributes: ' + str.join(', ', attrs) + '\n'
        return True, '\n' + txt
    return False, f'Variable {varname} not found'


def findvar(varname):
    ''' Find a variable and its parent object in the registered varlist set, based
        on varname, as passed by the stack.
        Variables can be searched in two ways:
        By name only: e.g., varname lat returns (traf, lat)
        By object: e.g., varname traf.lat returns (traf, lat)
        '''
    try:
        # Find a string matching 'a.b.c[d]', where everything except a is optional
        varset = re.findall(r'(\w+)(?<=.)*(?:\[(\w+)\])?', varname)
        # The actual variable is always the last
        name, index = varset[-1]
        # is a parent object passed? (e.g., traf.lat instead of just lat)
        if len(varset) > 1:
            obj = None
            # The first object should be in the varlist of Plot
            # As either a top-level object:
            if varset[0][0] in varlist:
                obj = varlist.get(varset[0][0])[0]
            else:
                for objname, objset in varlist.items():
                    if varset[0][0] in objset[1]:
                        obj = getattr(objset[0], varset[0][0])

            # Iterate over objectname,index pairs in varset
            for pair in varset[1:-1]:
                if obj is None:
                    break
                obj = getattr(obj, pair[0], None)

            if obj and hasattr(obj, name):
                return Variable(obj, varset[-2][0], name, index)
        else:
            # A parent object is not passed, we only have a variable name
            # this name should exist in Plot.vlist
            for objname, objset in varlist.items():
                if name in objset[1]:
                    return Variable(objset[0], objname, name, index)
    except:
        pass
    return None


class Variable:
    ''' Wrapper class for variable explorer.
        Keeps reference to parent object, parent name, and variable name. '''

    def __init__(self, parent, parentname, varname, index):
        self.parent = parent
        self.parentname = parentname
        self.varname = varname
        try:
            self.index = [int(i) for i in index]
        except ValueError:
            self.index = []

    def is_num(self):
        ''' py3 replacement of operator.isNumberType.'''
        v = getattr(self.parent, self.varname)
        return isinstance(v, Number) or \
            (isinstance(v, np.ndarray) and v.dtype.kind not in 'OSUV') or \
            (isinstance(v, Collection) and self.index and
             all([isinstance(v[i], Number) for i in self.index]))

    def get_type(self):
        ''' Return the a string containing the type name of this variable. '''
        return self.get().__class__.__name__

    def get(self):
        ''' Get a reference to the actual variable. '''
        if self.index:
            v = getattr(self.parent, self.varname)
            return [v[i] for i in self.index]
        return getproxied(getattr(self.parent, self.varname))



''' Conflict resolution base class. '''
import numpy as np

import bluesky as bs
from bluesky.core import Entity
from bluesky.stack import command
from bluesky.tools.aero import nm,ft


bs.settings.set_variable_defaults(asas_marh=1.01, asas_marv=1.01)


class ConflictResolution(Entity, replaceable=True):
    ''' Base class for Conflict Resolution implementations. '''
    def __init__(self):
        super().__init__()
        # [-] switch to activate priority rules for conflict resolution
        self.swprio = False  # switch priority on/off
        self.priocode = ''  # select priority mode
        self.resopairs = set()  # Resolved conflicts that are still before CPA

        # Resolution factors:
        # set < 1 to maneuver only a fraction of the resolution
        # set > 1 to add a margin to separation values
        self.resofach = bs.settings.asas_marh
        self.resofacv = bs.settings.asas_marv

        # Switches to guarantee last reso zone commands keep valid if cd zone changes
        self.resodhrelative = True # Size of resolution zone dh, vertically, set relative to CD zone
        self.resorrelative  = True # Size of resolution zone r, vertically, set relative to CD zone

        with self.settrafarrays():
            self.resooffac = np.array([], dtype=bool)
            self.noresoac = np.array([], dtype=bool)
            # whether the autopilot follows ASAS or not
            self.active = np.array([], dtype=bool)
            self.trk = np.array([])  # heading provided by the ASAS [deg]
            self.tas = np.array([])  # speed provided by the ASAS (eas) [m/s]
            self.alt = np.array([])  # alt provided by the ASAS [m]
            self.vs = np.array([])  # vspeed provided by the ASAS [m/s]

    def reset(self):
        super().reset()
        self.swprio = False
        self.priocode = ''
        self.resopairs.clear()
        self.resofach = bs.settings.asas_marh
        self.resofacv = bs.settings.asas_marv
        self.resodhrelative = True
        self.resorrelative  = True

    # By default all channels are controlled by self.active,
    # but they can be overloaded with separate variables or functions in a
    # derived ASAS Conflict Resolution class (@property decorator takes away
    # need for brackets when calling it so it can be overloaded by a variable)
    @property
    def hdgactive(self):
        ''' Return a boolean array sized according to the number of aircraft
            with True for all elements where heading is currently controlled by
            the conflict resolution algorithm.
        '''
        return self.active

    @property
    def vsactive(self):
        ''' Return a boolean array sized according to the number of aircraft
            with True for all elements where vertical speed is currently
            controlled by the conflict resolution algorithm.
        '''
        return self.active

    @property
    def altactive(self):
        ''' Return a boolean array sized according to the number of aircraft
            with True for all elements where altitude is currently controlled by
            the conflict resolution algorithm.
        '''
        return self.active

    @property
    def tasactive(self):
        ''' Return a boolean array sized according to the number of aircraft
            with True for all elements where speed is currently controlled by
            the conflict resolution algorithm.
        '''
        return self.active

    def resolve(self, conf, ownship, intruder):
        '''
            Resolve all current conflicts.
            This function should be reimplemented in a subclass for actual
            resolution of conflicts. See for instance
            bluesky.traffic.asas.mvp.
        '''
        # If resolution is off, and detection is on, and a conflict is detected
        # then asas will be active for that airplane. Since resolution is off, it
        # should then follow the auto pilot instructions.
        return ownship.ap.trk, ownship.ap.tas, ownship.ap.vs, ownship.ap.alt

    def update(self, conf, ownship, intruder):
        ''' Perform an update step of the Conflict Resolution implementation. '''
        if ConflictResolution.selected() is not ConflictResolution:
            # Only perform CR when an actual method is selected
            if conf.confpairs:
                self.trk, self.tas, self.vs, self.alt = self.resolve(conf, ownship, intruder)
            self.resumenav(conf, ownship, intruder)

    def resumenav(self, conf, ownship, intruder):
        '''
            Decide for each aircraft in the conflict list whether the ASAS
            should be followed or not, based on if the aircraft pairs passed
            their CPA.
        '''
        # Add new conflicts to resopairs and confpairs_all and new losses to lospairs_all
        self.resopairs.update(conf.confpairs)

        # Conflict pairs to be deleted
        delpairs = set()
        changeactive = dict()

        # smallest relative angle between vectors of heading a and b
        def anglediff(a, b):
            d = a - b
            if d > 180:
                return anglediff(a, b + 360)
            elif d < -180:
                return anglediff(a + 360, b)
            else:
                return d
            

        # Look at all conflicts, also the ones that are solved but CPA is yet to come
        for conflict in self.resopairs:
            idx1, idx2 = bs.traf.id2idx(conflict)
            # If the ownship aircraft is deleted remove its conflict from the list
            if idx1 < 0:
                delpairs.add(conflict)
                continue

            if idx2 >= 0:
                # Distance vector using flat earth approximation
                re = 6371000.
                dist = re * np.array([np.radians(intruder.lon[idx2] - ownship.lon[idx1]) *
                                      np.cos(0.5 * np.radians(intruder.lat[idx2] +
                                                              ownship.lat[idx1])),
                                      np.radians(intruder.lat[idx2] - ownship.lat[idx1])])

                # Relative velocity vector
                vrel = np.array([intruder.gseast[idx2] - ownship.gseast[idx1],
                                 intruder.gsnorth[idx2] - ownship.gsnorth[idx1]])

                # Check if conflict is past CPA
                past_cpa = np.dot(dist, vrel) > 0.0

                rpz = np.max(conf.rpz[[idx1, idx2]])
                # hor_los:
                # Aircraft should continue to resolve until there is no horizontal
                # LOS. This is particularly relevant when vertical resolutions
                # are used.
                hdist = np.linalg.norm(dist)
                hor_los = hdist < rpz

                # Bouncing conflicts:
                # If two aircraft are getting in and out of conflict continously,
                # then they it is a bouncing conflict. ASAS should stay active until
                # the bouncing stops.
                is_bouncing = \
                    abs(anglediff(ownship.trk[idx1], intruder.trk[idx2])) < 30.0 and \
                    hdist < rpz * self.resofach

            # Start recovery for ownship if intruder is deleted, or if past CPA
            # and not in horizontal LOS or a bouncing conflict
            if idx2 >= 0 and (not past_cpa or hor_los or is_bouncing):
                # Enable ASAS for this aircraft
                changeactive[idx1] = True
            else:
                # Switch ASAS off for ownship if there are no other conflicts
                # that this aircraft is involved in.
                changeactive[idx1] = changeactive.get(idx1, False)
                # If conflict is solved, remove it from the resopairs list
                delpairs.add(conflict)

        for idx, active in changeactive.items():
            # Loop a second time: this is to avoid that ASAS resolution is
            # turned off for an aircraft that is involved simultaneously in
            # multiple conflicts, where the first, but not all conflicts are
            # resolved.
            self.active[idx] = active
            if not active:
                # Waypoint recovery after conflict: Find the next active waypoint
                # and send the aircraft to that waypoint.
                iwpid = bs.traf.ap.route[idx].findact(idx)
                if iwpid != -1:  # To avoid problems if there are no waypoints
                    bs.traf.ap.route[idx].direct(
                        idx, bs.traf.ap.route[idx].wpname[iwpid])

        # Remove pairs from the list that are past CPA or have deleted aircraft
        self.resopairs -= delpairs

    @command(name='PRIORULES')
    def setprio(self, flag : bool = None, priocode=''):
        ''' Define priority rules (right of way) for conflict resolution. '''
        if flag is None:
            if self.__class__ is ConflictResolution:
                return False, 'No conflict resolution enabled.'
            return False, f'Resolution algorithm {self.__class__.name} hasn\'t implemented priority.'

        self.swprio = flag
        self.priocode = priocode
        return True

    @command(name='NORESO')
    def setnoreso(self, *idx : 'acid'):
        ''' ADD or Remove aircraft that nobody will avoid.
        Multiple aircraft can be sent to this function at once. '''
        if not idx:
            return True, 'NORESO [ACID, ... ] OR NORESO [GROUPID]' + \
                         '\nCurrent list of aircraft nobody will avoid:' + \
                         ', '.join(np.array(bs.traf.id)[self.noresoac])
        idx = list(idx)
        self.noresoac[idx] = np.logical_not(self.noresoac[idx])
        return True

    @command(name='RESOOFF')
    def setresooff(self, *idx : 'acid'):
        ''' ADD or Remove aircraft that will not avoid anybody else.
            Multiple aircraft can be sent to this function at once. '''
        if not idx:
            return True, 'NORESO [ACID, ... ] OR NORESO [GROUPID]' + \
                         '\nCurrent list of aircraft will not avoid anybody:' + \
                         ', '.join(np.array(bs.traf.id)[self.resooffac])
        else:
            idx = list(idx)
            self.resooffac[idx] = np.logical_not(self.resooffac[idx])
            return True

    @command(name='RFACH', aliases=('RESOFACH', 'HRFAC', 'HRESOFAC'))
    def setresofach(self, factor : float = None):
        ''' Set resolution factor horizontal
            (to maneuver only a fraction of a resolution vector)
        '''
        if factor is None:
            return True, f'RFACH [FACTOR]\nCurrent horizontal resolution factor is: {self.resofach}'
        else:
            self.resofach = factor
            self.resorrelative = True  # Size of resolution zone r, vertically, set relative to CD zone
            return True, f'Horizontal resolution factor set to {self.resofach}'

    @command(name='RFACV', aliases=('RESOFACV',))
    def setresofacv(self, factor: float = None):
        ''' Set resolution factor vertical (to maneuver only a fraction of a resolution vector). '''
        if factor is None:
            return True, f'RFACV [FACTOR]\nCurrent vertical resolution factor is: {self.resofacv}'
        self.resofacv = factor
        # Size of resolution zone dh, vertically, set relative to CD zone
        self.resodhrelative = True
        return True, f'Vertical resolution factor set to {self.resofacv}'

    @command(name='RSZONER', aliases=('RESOZONER',))
    def setresozoner(self, zoner : float = None):
        ''' Set resolution factor horizontal, but then with absolute value
            (to maneuver only a fraction of a resolution vector)
        '''
        if not bs.traf.cd.global_rpz:
            self.resorrelative = True
            return False, 'RSZONER [radiusnm]\nCan only set resolution factor when simulation contains aircraft with different RPZ,\nUse RFACH instead.'
        if zoner is None:
            return True, f'RSZONER [radiusnm]\nCurrent horizontal resolution factor is: {self.resofach}, resulting in radius: {self.resofach*bs.traf.cd.rpz_def/nm} nm'

        self.resofach = zoner / bs.traf.cd.rpz_def * nm
        # Size of resolution zone r, vertically, no longer relative to CD zone
        self.resorrelative = False
        return True, f'Horizontal resolution factor updated to {self.resofach}, resulting in radius: {zoner} nm'

    @command(name='RSZONEDH', aliases=('RESOZONEDH',))
    def setresozonedh(self, zonedh : float = None):
        '''
        Set resolution factor vertical (to maneuver only a fraction of a resolution vector),
        but then with absolute value
        '''
        if not bs.traf.cd.global_hpz:
            self.resodhrelative = True
            return False, 'RSZONEH [zonedhft]\nCan only set resolution factor when simulation contains aircraft with different HPZ,\nUse RFACV instead.'
        if zonedh is None:
            return True, f'RSZONEDH [zonedhft]\nCurrent vertical resolution factor is: {self.resofacv}, resulting in height: {self.resofacv*bs.traf.cd.hpz_def/ft} ft'

        self.resofacv = zonedh / bs.traf.cd.hpz_def * ft
        # Size of resolution zone dh, vertically, no longer relative to CD zone
        self.resodhrelative = False
        return True, f'Vertical resolution factor updated to {self.resofacv}, resulting in height: {zonedh} ft'

    @staticmethod
    @command(name='RESO')
    def setmethod(name : 'txt' = ''):
        ''' Select a Conflict Resolution method. '''
        # Get a dict of all registered CR methods
        methods = ConflictResolution.derived()
        names = ['OFF' if n == 'CONFLICTRESOLUTION' else n for n in methods]

        if not name:
            curname = 'OFF' if ConflictResolution.selected() is ConflictResolution \
                else ConflictResolution.selected().__name__
            return True, f'Current CR method: {curname}' + \
                         f'\nAvailable CR methods: {", ".join(names)}'
        # Check if the requested method exists
        if name == 'OFF':
            ConflictResolution.select()
            return True, 'Conflict Resolution turned off.'
        method = methods.get(name, None)
        if method is None:
            return False, f'{name} doesn\'t exist.\n' + \
                          f'Available CR methods: {", ".join(names)}'

        # Select the requested method
        method.select()
        return True, f'Selected {method.__name__} as CR method.'
